import unittest
import flavio
from wilson import Wilson
from .Vllgamma import *

### implement test
class TestVllgamma(unittest.TestCase):
    def test_np(self):
        wc,br=Wilson({'CVRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),8.3949e-6
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->muegamma)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->muegamma)',wc),flavio.np_prediction('BR(J/psi->muegamma)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)

        wc,br=Wilson({'CSRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),6.2935e-6
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->muegamma)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->muegamma)',wc),flavio.np_prediction('BR(J/psi->muegamma)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)

        wc,br=Wilson({'CVRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),1.2887e-6
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->tauegamma)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->tauegamma)',wc),flavio.np_prediction('BR(J/psi->tauegamma)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)
        
        wc,br=Wilson({'CSRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),9.1097e-7
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->tauegamma)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->tauegamma)',wc),flavio.np_prediction('BR(J/psi->tauegamma)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)

        


import unittest
import flavio
from wilson import Wilson
from .Pll import *

### implement test
class TestPll(unittest.TestCase):
    def test_np(self):
        wc,br=Wilson({'CVRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),7.7348e-8
        self.assertAlmostEqual(flavio.np_prediction('BR(eta_c(1S)->mue)',wc), br,delta=0.001*br)

        wc,br=Wilson({'CSRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),8.4734e-5
        self.assertAlmostEqual(flavio.np_prediction('BR(eta_c(1S)->mue)',wc), br,delta=0.001*br)

        wc,br=Wilson({'CVRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),9.1346e-6
        self.assertAlmostEqual(flavio.np_prediction('BR(eta_c(1S)->taue)',wc), br,delta=0.001*br)
        
        wc,br=Wilson({'CSRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),3.5384e-5
        self.assertAlmostEqual(flavio.np_prediction('BR(eta_c(1S)->taue)',wc), br,delta=0.001*br)



from math import sqrt, exp
import warnings
import flavio

def omega_fct(q2, mLb, mLst):
    # eq. (73) in arXiv:2009.09313v1
    return (mLb*mLb + mLst*mLst - q2)/(2*mLb*mLst)

def ff_formula(F, A, omega):
    # eq. (75)
    return F + A*(omega - 1)

_process_dict = {}
_process_dict['Lambdab->Lambda(1520)'] = {'X': 'Lambda(1520)'}

def formfactors(process, par, q2):
    r"Formfactors for $\Lambda_b\to L(1520)$ from Lattice QCD as in arXiv:2009.09313v1"
    flavio.citations.register("Meinel:2020owd")

    pd = _process_dict[process]
    mL = par['m_'+pd['X']]
    mLb = par['m_Lambdab']

    omega = omega_fct(q2, mLb, mL)
    ff_list = ['f0', 'fplus', 'fperp', 'fperpPrim',
               'g0', 'gplus', 'gperp', 'gperpPrim',
               'hplus', 'hperp', 'hperpPrim', 'hTplus', 'hTperp', 'hTperpPrim']

    ff_dict = {}
    for e in ff_list:
        F = par[process+' '+e+' F']
        A = par[process+' '+e+' A']

        ff_dict[e] = ff_formula(F, A, omega)

    return ff_dict, mL, mLb


def ff_equiv(process, q2, par):
    # eq. (A21) - (A34) and (6)
    ff_dict, mLst, mLb = formfactors(process, par, q2)
    splus  = (mLb + mLst)**2 - q2
    sminus = (mLb - mLst)**2 - q2

    if q2 < 16.0 or q2 > (mLb - mLst)**2:
        warnings.warn(f'Lattice QCD form factors are used out of the allowed q2 region [16.0; {(mLb - mLst)**2:.3}].')

    ff = {}
    ff['fVt']    = ( mLst/splus )*ff_dict['f0']
    ff['fV0']    = ( mLst/sminus )*ff_dict['fplus']
    ff['fVperp'] = ( mLst/sminus )*ff_dict['fperp']
    ff['fVg']    = ff_dict['fperpPrim']

    ff['fAt']    = ( mLst/sminus )*ff_dict['g0']
    ff['fA0']    = ( mLst/splus )*ff_dict['gplus']
    ff['fAperp'] = ( mLst/splus )*ff_dict['gperp']
    ff['fAg']    = -ff_dict['gperpPrim']

    ff['fTt']    = 0
    ff['fT0']    = ( mLst/sminus )*ff_dict['hplus']
    ff['fTperp'] = ( mLst/sminus )*ff_dict['hperp']
    ff['fTg']    = ( mLb + mLst )*ff_dict['hperpPrim']

    ff['fT5t']   = 0
    ff['fT50']   = ( mLst/splus )*ff_dict['hTplus']
    ff['fT5perp']= ( mLst/splus )*ff_dict['hTperp']
    ff['fT5g']   = -( mLb - mLst )*ff_dict['hTperpPrim']

    return ff



import unittest
import flavio
from flavio.physics.dileptons import pplnu
from wilson import wcxf
from flavio.physics.dileptons.test_ppll import uses_pdf
import numpy as np
from flavio.config import config

par = flavio.default_parameters.get_central_all()
par2 = par.copy()
par2['s2w'] = 0.22137905667111532  # correction factor to match the madgraph input scheme
GeVtopb = 0.3894*10**9

class Test_QQLNU(unittest.TestCase):
    @uses_pdf
    def test_sm_spectrum(self):
        r"""Compare the SM $m_T$ spectrum to madgraph events in 10 bins from 200 GeV to 2 TeV
            MG events simulated with NNPDF30_nnlo_as_0118
            2% acceptable error"""

        spectrum_mg = np.asarray([3.41707366e-02, 1.02592426e-02, 3.20641729e-03, 9.63481603e-04,
                                    2.81233386e-04, 8.12019322e-05, 2.13711295e-05, 5.30226309e-06,
                                    1.14687576e-06])
        # Number of SM events generated in MG [66095., 25637., 33458., 48654., 18351.,  6849., 59869., 32043., 9044.]

        s = 13e3**2
        logbins = np.linspace(np.log10(200),np.log10(2000),10)
        bins = 10**logbins
        nbins = len(bins)-1
        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            width = bins[i+1]-bins[i]
            spectrum = pplnu.sigma_qqlnu_int(s, bins[i], bins[i+1], 'mu', 0, par2, center**2, 0, newphys=False)*GeVtopb/width
            err = (spectrum-spectrum_mg[i])/spectrum_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqlnu_sm(self):
        """Test the SM pediction for the R ratio from 1 to 2 TeV"""
        R = flavio.sm_prediction('R_13(pp->munu)', 1e3, 2e3)
        self.assertEqual(R, 1,msg=f'SM prediction for R ratio: {R} (has to be 1)')

    @uses_pdf
    def test_qqlnu_np_sm(self):
        r"""Test the R ratio with presence of wilson coefficients set to zero"""
        wc_obj = flavio.WilsonCoefficients()
        R = pplnu.R_sigma_qqlnu_int(13e3**2, 1300, 1800, 'mu', wc_obj, par)
        self.assertEqual(R, 1,msg=f'SM prediction for R ratio: {R} (has to be 1)')

    @uses_pdf
    def test_qqlnu_np_vector(self):
        r"""Test the R ratio for one specific vector WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118 and SMEFTsim_general_MwScheme_UFO model
            2% acceptable error"""

        R_mg = np.asarray([ 8.76973438, 12.82963675, 18.01933393])
        # Number of NP events generated in MG [25183, 16648, 11444]
        # Number of SM events generated in MG [23536, 10207, 4851]

        bins = np.asarray([1200., 1400., 1600., 1800.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'lq3_2211': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = pplnu.R_sigma_qqlnu_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqlnu_np_scalar(self):
        r"""Test the R ratio for one specific scalar WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118 and SMEFTsim_general_MwScheme_UFO model
            2% acceptable error"""

        R_mg = np.asarray([1.00938347, 1.01333147, 1.01762706])
        # Number of NP events generated in MG [25603, 15708, 9833]
        # Number of SM events generated in MG [23536, 10207, 4851]

        bins = np.asarray([1200., 1400., 1600., 1800.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'lequ1_2232': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = pplnu.R_sigma_qqlnu_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqlnu_np_tensor(self):
        r"""Test the R ratio for one specific tensor WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118 and SMEFTsim_general_MwScheme_UFO model
            2% acceptable error"""

        R_mg = np.asarray([1.42841172, 1.63238727, 1.87684397])
        # Number of NP events generated in MG [25779, 16499, 10905]
        # Number of SM events generated in MG [23536, 10207, 4851]

        bins = np.asarray([1200., 1400., 1600., 1800.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'lequ3_2212': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = pplnu.R_sigma_qqlnu_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')



import unittest
import numpy as np
import flavio
from flavio.util import get_datapath
from wilson import Wilson

path = get_datapath('flavio', 'data/test/')
sm_array = np.load(path+'2009.09313_digitized.npz')

class TestLambdabLambda1520(unittest.TestCase):

    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        self.wc_np_mu = Wilson({'C9_bsmumu': -1.11}, scale=4.8, eft='WET', basis='flavio')
        self.ang_coef_List = ['1c', '1cc', '1ss', '2c', '2cc', '2ss', '3ss', '4ss', '5s', '5sc', '6s', '6sc']

    def assert_obs(self, name, target, NP=False, rtol=0.01, scalef=1, **kwargs):
        if NP:
            c = flavio.np_prediction(name, self.wc_np_mu, **kwargs)*scalef
        else:
            c = flavio.sm_prediction(name, **kwargs)*scalef
        self.assertAlmostEqual(c/target, 1, delta=rtol)
        return None

    def assert_sm_arrays(self, name, targetArray, bound, rtol=0.01, atol=0, scalef=1):
        values_to_use = targetArray[0] < 16.79
        q2_vals = targetArray[0][values_to_use]
        target = targetArray[1][values_to_use]
        c = np.array([flavio.sm_prediction(name, q2=q2val)*scalef for q2val in q2_vals])
        if bound == 'central':
            shift = 0
        else:
            unc = np.array([flavio.sm_uncertainty(name, q2=q2val, N=100)*scalef for q2val in q2_vals])
            if bound == 'upper':
                shift = +unc
            elif bound == 'lower':
                shift = -unc
        np.testing.assert_allclose(c+shift, target, rtol=rtol, atol=atol, verbose=True)
        return None


class TestLambdabLambda1520_FF(TestLambdabLambda1520):

    def test_lambdab_lambda1520_ff(self):
        flavio.config['implementation']['Lambdab->Lambda(1520) form factor'] = 'Lambdab->Lambda(1520) LatticeQCD'
        br = flavio.default_parameters.get_central('BR(Lambda(1520)->NKbar)_exp')
        br_inv2 = 2e9/br

        # Comparison to figure 6 in arXiv:2009.09313v2
        # Scale factor BRinv2 used since in np-array factor e^(-9) not included
        dBR = 'dBR/dq2(Lambdab->Lambda(1520)mumu)'
        self.assert_sm_arrays(dBR, sm_array['dB_central'], 'central', scalef=br_inv2, rtol=0.02, atol=0.03)
        self.assert_sm_arrays(dBR, sm_array['dB_upper'], 'upper', scalef=br_inv2, rtol=0.08, atol=0.03)
        self.assert_sm_arrays(dBR, sm_array['dB_lower'], 'lower', scalef=br_inv2, rtol=0.08, atol=0.03)

        # Comparison to figure 9 right in arXiv:2009.09313v2
        Afb = 'AFBl(Lambdab->Lambda(1520)mumu)'
        self.assert_sm_arrays(Afb, sm_array['AFBl_central'], 'central', rtol=0.02, atol=0.03)
        self.assert_sm_arrays(Afb, sm_array['AFBl_upper'], 'upper', rtol=0.08, atol=0.03)
        self.assert_sm_arrays(Afb, sm_array['AFBl_lower'], 'lower', rtol=0.08, atol=0.03)

        # Comparison to figure 7 top right in arXiv:2009.09313v2
        S1cc = 'S_1cc(Lambdab->Lambda(1520)mumu)'
        self.assert_sm_arrays(S1cc, sm_array['S1cc_central'], 'central', rtol=0.02, atol=0.03)
        self.assert_sm_arrays(S1cc, sm_array['S1cc_upper'], 'upper', rtol=0.08, atol=0.03)
        self.assert_sm_arrays(S1cc, sm_array['S1cc_lower'], 'lower', rtol=0.08, atol=0.03)

        flavio.config['implementation']['Lambdab->Lambda(1520) form factor'] = 'Lambdab->Lambda(1520) MCN'
        # Comparison to figure on slide 12 S.Meinel b-baryon FEST 2020
        self.assert_sm_arrays(dBR, sm_array['dB_QM'], 'central', scalef=br_inv2, rtol=0.02, atol=0.03)

        # Comparison to figure on slide 14 S.Meinel b-baryon FEST 2020
        self.assert_sm_arrays(Afb, sm_array['AFBl_QM'], 'central', rtol=0.02, atol=0.03)

        # Comparison to figure on slide 13 S.Meinel b-baryon FEST 2020
        self.assert_sm_arrays(S1cc, sm_array['S1cc_QM'], 'central', rtol=0.02, atol=0.03)


class TestLambdabLambda1520ll(TestLambdabLambda1520):

    def test_lambdablambda1520ll_binned_qmff_sm(self):
        # compare to SM values in table 1 of 2005.09602 using the quark model form factors
        # Differences due to C9eff = C9 and C7eff = C7
        flavio.config['implementation']['Lambdab->Lambda(1520) form factor'] = 'Lambdab->Lambda(1520) MCN'

        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 0.397, rtol=0.18, scalef=1e9, q2min=0.1, q2max=3)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 1.29, rtol=0.08, scalef=1e9, q2min=3, q2max=6)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 3.22, rtol=0.10, scalef=1e9, q2min=6, q2max=8.68)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 0.95, rtol=0.08, scalef=1e9, q2min=1, q2max=6)

        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', 0.048, rtol=0.57, q2min=0.1, q2max=3)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.127, rtol=0.16, q2min=3, q2max=6)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.235, rtol=0.03, q2min=6, q2max=8.68)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.098, rtol=0.24, q2min=1, q2max=6)


        self.assert_obs('<S_1cc>(Lambdab->Lambda(1520)mumu)', 0.181, rtol=0.29, q2min=0.1, q2max=3)
        self.assert_obs('<S_1cc>(Lambdab->Lambda(1520)mumu)', 0.242, rtol=0.07, q2min=3, q2max=6)
        self.assert_obs('<S_1cc>(Lambdab->Lambda(1520)mumu)', 0.361, rtol=0.05, q2min=6, q2max=8.68)
        self.assert_obs('<S_1cc>(Lambdab->Lambda(1520)mumu)', 0.221, rtol=0.06, q2min=1, q2max=6)

    def test_lambdablambda1520ll_binned_qmff_np(self):
        # compare to NP values in table 1 of 2005.09602 using quark model form factors
        flavio.config['implementation']['Lambdab->Lambda(1520) form factor'] = 'Lambdab->Lambda(1520) MCN'

        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 0.337, NP=True, rtol=0.19, scalef=1e9, q2min=0.1, q2max=3)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 1.04, NP=True, rtol=0.06, scalef=1e9, q2min=3, q2max=6)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 2.58, NP=True, rtol=0.07, scalef=1e9, q2min=6, q2max=8.68)
        self.assert_obs('<dBR/dq2>(Lambdab->Lambda(1520)mumu)', 0.77, NP=True, rtol=0.07, scalef=1e9, q2min=1, q2max=6)

        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', 0.098  , NP=True, rtol=0.26, q2min=0.1, q2max=3)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.059, NP=True, rtol=0.38, q2min=3, q2max=6)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.166, NP=True, rtol=0.01, q2min=6, q2max=8.68)
        self.assert_obs('<AFBl>(Lambdab->Lambda(1520)mumu)', -0.031, NP=True, rtol=0.84, q2min=1, q2max=6)

    def test_lambdablambda1520mm_all_observables(self):
        # Test if all the observables with l=mu work fine.
        names = [
            '<dBR/dq2>(Lambdab->Lambda(1520)mumu)',
            '<AFBh>(Lambdab->Lambda(1520)mumu)',
            '<AFBlh>(Lambdab->Lambda(1520)mumu)',
            '<FL>(Lambdab->Lambda(1520)mumu)',
            '<A_1c>(Lambdab->Lambda(1520)mumu)',
            '<S_1c>(Lambdab->Lambda(1520)mumu)',
            '<A_1cc>(Lambdab->Lambda(1520)mumu)',
            '<S_1cc>(Lambdab->Lambda(1520)mumu)',
            '<A_1ss>(Lambdab->Lambda(1520)mumu)',
            '<S_1ss>(Lambdab->Lambda(1520)mumu)',
            '<A_2c>(Lambdab->Lambda(1520)mumu)',
            '<S_2c>(Lambdab->Lambda(1520)mumu)',
            '<A_2cc>(Lambdab->Lambda(1520)mumu)',
            '<S_2cc>(Lambdab->Lambda(1520)mumu)',
            '<A_2ss>(Lambdab->Lambda(1520)mumu)',
            '<S_2ss>(Lambdab->Lambda(1520)mumu)',
            '<A_3ss>(Lambdab->Lambda(1520)mumu)',
            '<S_3ss>(Lambdab->Lambda(1520)mumu)',
            '<A_4ss>(Lambdab->Lambda(1520)mumu)',
            '<S_4ss>(Lambdab->Lambda(1520)mumu)',
            '<A_5sc>(Lambdab->Lambda(1520)mumu)',
            '<S_5sc>(Lambdab->Lambda(1520)mumu)',
            '<A_6s>(Lambdab->Lambda(1520)mumu)',
            '<S_6s>(Lambdab->Lambda(1520)mumu)',
            '<A_6sc>(Lambdab->Lambda(1520)mumu)',
            '<S_6sc>(Lambdab->Lambda(1520)mumu)',
        ]
        for name in names:
            flavio.sm_prediction(name, q2min=1, q2max=6)

    def test_lambdablambda1520ee_all_observables(self):
        # Test if all the observables with l=e work fine.
        names = [
            '<dBR/dq2>(Lambdab->Lambda(1520)ee)',
            '<AFBh>(Lambdab->Lambda(1520)ee)',
            '<AFBlh>(Lambdab->Lambda(1520)ee)',
            '<FL>(Lambdab->Lambda(1520)ee)',
            '<A_1c>(Lambdab->Lambda(1520)ee)',
            '<S_1c>(Lambdab->Lambda(1520)ee)',
            '<A_1cc>(Lambdab->Lambda(1520)ee)',
            '<S_1cc>(Lambdab->Lambda(1520)ee)',
            '<A_1ss>(Lambdab->Lambda(1520)ee)',
            '<S_1ss>(Lambdab->Lambda(1520)ee)',
            '<A_2c>(Lambdab->Lambda(1520)ee)',
            '<S_2c>(Lambdab->Lambda(1520)ee)',
            '<A_2cc>(Lambdab->Lambda(1520)ee)',
            '<S_2cc>(Lambdab->Lambda(1520)ee)',
            '<A_2ss>(Lambdab->Lambda(1520)ee)',
            '<S_2ss>(Lambdab->Lambda(1520)ee)',
            '<A_3ss>(Lambdab->Lambda(1520)ee)',
            '<S_3ss>(Lambdab->Lambda(1520)ee)',
            '<A_4ss>(Lambdab->Lambda(1520)ee)',
            '<S_4ss>(Lambdab->Lambda(1520)ee)',
            '<A_5sc>(Lambdab->Lambda(1520)ee)',
            '<S_5sc>(Lambdab->Lambda(1520)ee)',
            '<A_6s>(Lambdab->Lambda(1520)ee)',
            '<S_6s>(Lambdab->Lambda(1520)ee)',
            '<A_6sc>(Lambdab->Lambda(1520)ee)',
            '<S_6sc>(Lambdab->Lambda(1520)ee)',
        ]
        for name in names:
            flavio.sm_prediction(name, q2min=1, q2max=6)



r"""$S\to ll^\prime$ branching ratio"""

from flavio.classes import Observable, Prediction
import flavio
from flavio.physics.running import running
from flavio.physics.common import lambda_K
from flavio.physics.quarkonium.Vll import wc_sector
import numpy as np

meson_quark = { 'chi_c0(1P)' : 'cc', 
                'chi_b0(1P)': 'bb',
                'chi_b0(2P)': 'bb',
                }


def getS_lfv(wc_obj,par,S,l1,l2,CeGGij,CeGGji):
    # renormalization scale
    scale = flavio.config['renormalization scale'][S]
    alphas = running.get_alpha_s(par, scale)
    # Wilson coefficients
    wc = wc_obj.get_wc(wc_sector[(l1,l2)], scale, par)

    mS = par['m_'+S]   
    # The form factor in the parameters_uncorrelated.yml file follows 1607.00815. Thus we have to include an additional factor of "-i" which drops out in the absence of an anomaly term.
    fS=par['f_'+S]*(-1j) 
    aS=par['a_'+S]
    qq=meson_quark[S]
    aPfac=1j*aS*4.*np.pi/alphas
    SR= aPfac * CeGGij +  (mS*fS/2.)*(wc['CSRR_'+l1+l2+qq]+wc['CSRL_'+l1+l2+qq])  
    SL= aPfac * CeGGji + (mS*fS/2.)*(wc['CSRL_'+l2+l1+qq]+wc['CSRR_'+l2+l1+qq]).conjugate() 
    return SL,SR


def Sll_br(wc_obj, par,S, l1,l2,CeGGij,CeGGji):
    r"""Branching ratio for the lepton-flavour violating leptonic decay P -> l l' """
    #####branching ratio obtained from 2207.10913#####
    flavio.citations.register("Calibbi:2022ddo")
    mP = par['m_'+S]   
    tauP = par['tau_'+S]  
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    y1=ml1/mP
    y2=ml2/mP
    y1s=y1**2
    y2s=y2**2
    SL,SR = getS_lfv(wc_obj,par,S,l1,l2,CeGGij,CeGGji)
    return  tauP*mP/(16.*np.pi) * np.sqrt(lambda_K(1,y1s,y2s)) * ((1-y1s-y2s)*(np.abs(SL)**2+np.abs(SR)**2) -4*y1*y2 *(SL*SR.conjugate()).real)


def Sll_br_func(S,  l1, l2):
    def fct(wc_obj, par,CeGGij=0,CeGGji=0):
        return Sll_br(wc_obj, par, S,  l1, l2,CeGGij,CeGGji)
    return fct


def Sll_br_comb_func(S,  l1, l2):
    def fct(wc_obj, par,CeGGij=0,CeGGji=0):
        return Sll_br(wc_obj, par, S,  l1, l2,CeGGij,CeGGji)+ Sll_br(wc_obj, par, S,  l2, l1,CeGGij,CeGGji)
    return fct


# Observable and Prediction instances
_hadr = { 
    'chi_c0(1P)': {'tex': r"\chi_{c0}(1P)\to", 'S': 'chi_c0(1P)', 'Q': 2./3., },
    'chi_b0(1P)': {'tex': r"\chi_{b0}(1P)\to", 'S': 'chi_b0(1P)', 'Q': -1./3., },
    'chi_b0(2P)': {'tex': r"\chi_{b0}(2P)\to", 'S': 'chi_b0(2P)', 'Q': -1./3., },
 }

_tex = {'ee': r'e^+e^-', 'mumu': r'\mu^+\mu^-', 'tautau': r'\tau^+\tau^-', 
    'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-',
    'emu,mue': r'e^\pm\mu^\mp', 'etau,taue': r'e^\pm\tau^\mp','mutau,taumu': r'\mu^\pm\tau^\mp',
    'mue,emu': r'e^\pm\mu^\mp', 'taue,etau': r'e^\pm\tau^\mp','taumu,mutau': r'\mu^\pm\tau^\mp'
    }


def _define_obs_S_ll(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $S\to \ell^+\ell^-$ :: $' + _process_tex + r"$"
    _obs_name = "BR("+_hadr[M]['S']+"->"+''.join(ll)+")"
    _obs = Observable(_obs_name,arguments=["CeGGij","CeGGji"])
    _obs.set_description(r"Branching ratio of $"+_process_tex+r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name


for M in _hadr:
    for (l1,l2) in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
        _obs_name = _define_obs_S_ll(M, (l1,l2))
        Prediction(_obs_name, Sll_br_func(_hadr[M]['S'], l1,l2))
    for ll in [('e','mu'), ('e','tau'), ('mu','tau')]:
        # Combined l1+ l2- + l2+ l1- lepton flavour violating decays
        _obs_name = _define_obs_S_ll(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Sll_br_comb_func(_hadr[M]['S'],  l1,l2))
        _obs_name = _define_obs_S_ll(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Sll_br_comb_func(_hadr[M]['S'],  l1,l2))
 



r"""Functions for $\Lambda_b \to \Lambda(1520)(\to NK) \ell^+ \ell⁻$ decays as in arXiv:1903.00448."""

import flavio
from math import sqrt, pi
from flavio.physics.bdecays.common import lambda_K, beta_l, meson_quark, meson_ff
from flavio.classes import Observable, Prediction, AuxiliaryQuantity
from flavio.physics.common import conjugate_par, conjugate_wc, add_dict
import warnings


def helicity_amps(q2, mLb, mL, ff):
    flavio.citations.register("Descotes-Genon:2019dbw")

    # Hadronic helicity amplitudes of Lb->L(1520)(->pK)l+l-
    sp = (mLb + mL)**2 - q2
    sm = (mLb - mL)**2 - q2

    H = {}
    # non-vanishing amplitudes in the vector part (V)
    # eqs (3.8) of arXiv:1903.00448
    # 1 for 1/2, 3 for 3/2
    H['tV+1+1'] = ff['fVt'] * (mLb - mL)/sqrt(q2) * (sp * sqrt(sm))/(sqrt(6) * mL)
    H['tV-1-1'] = H['tV+1+1']
    H['0V+1+1'] = -ff['fV0'] * (mLb + mL)/sqrt(q2) * (sm * sqrt(sp))/(sqrt(6) * mL)
    H['0V-1-1'] = H['0V+1+1']
    H['+V+1-1'] = -ff['fVperp'] * (sm * sqrt(sp))/(sqrt(3) * mL)
    H['-V-1+1'] = H['+V+1-1']
    H['+V-1-3'] = ff['fVg'] * sqrt(sp)
    H['-V+1+3'] = H['+V-1-3']

    # for the axial part (A), eqs (3.9)
    H['tA+1+1'] = ff['fAt'] * (mLb + mL)/sqrt(q2) * (sm * sqrt(sp))/(sqrt(6) * mL)
    H['tA-1-1'] = -H['tA+1+1']
    H['0A+1+1'] = -ff['fA0'] * (mLb - mL)/sqrt(q2) * (sp * sqrt(sm))/(sqrt(6) * mL)
    H['0A-1-1'] = -H['0A+1+1']
    H['+A+1-1'] = ff['fAperp'] * (sp * sqrt(sm))/(sqrt(3) * mL)
    H['-A-1+1'] = -H['+A+1-1']
    H['+A-1-3'] = -ff['fAg'] * sqrt(sm)
    H['-A+1+3'] = -H['+A-1-3']

    # for the tensor part (T), eqs (3.15)
    H['0T+1+1'] = ff['fT0'] * sqrt(q2) * (sm * sqrt(sp))/(sqrt(6) * mL)
    H['0T-1-1'] = H['0T+1+1']
    H['+T+1-1'] = ff['fTperp'] * (mLb + mL) * (sm * sqrt(sp))/(sqrt(3) * mL)
    H['-T-1+1'] = H['+T+1-1']
    H['+T-1-3'] = -ff['fTg'] * sqrt(sp)
    H['-T+1+3'] = H['+T-1-3']

    H['0T5+1+1'] = -ff['fT50'] * sqrt(q2) * (sp * sqrt(sm))/(sqrt(6) * mL)
    H['0T5-1-1'] = -H['0T5+1+1']
    H['+T5+1-1'] = ff['fT5perp'] * (mLb - mL) * (sp * sqrt(sm))/(sqrt(3) * mL)
    H['-T5-1+1'] = -H['+T5+1-1']
    H['+T5-1-3'] = -ff['fT5g'] * sqrt(sm)
    H['-T5+1+3'] = -H['+T5-1-3']

    return H


def transversity_amps(ha, q2, mLb, mL, mqh, wc, prefactor, t):
    # Hadronic transversity amplitudes
    # defined as in eqs (3.18) and (3.20)

    ## by taking wc_eff
    C910Lpl = (wc['v'] - wc['a']) + (wc['vp'] - wc['ap'])
    C910Rpl = (wc['v'] + wc['a']) + (wc['vp'] + wc['ap'])
    C910Lmi = (wc['v'] - wc['a']) - (wc['vp'] - wc['ap'])
    C910Rmi = (wc['v'] + wc['a']) - (wc['vp'] + wc['ap'])

    A = {}

    A['Bperp1', 'L'] = sqrt(2)*( C910Lpl * ha['+V-1-3'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['+T-1-3'])
    A['Bperp1', 'R'] = sqrt(2)*( C910Rpl * ha['+V-1-3'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['+T-1-3'])
    A['Bpara1', 'L'] = -sqrt(2)*( C910Lmi * ha['+A-1-3'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['+T5-1-3'])
    A['Bpara1', 'R'] = -sqrt(2)*( C910Rmi * ha['+A-1-3'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['+T5-1-3'])


    A['Aperp1', 'L'] = sqrt(2)*( C910Lpl * ha['+V+1-1'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['+T+1-1'])
    A['Aperp1', 'R'] = sqrt(2)*( C910Rpl * ha['+V+1-1'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['+T+1-1'])
    A['Apara1', 'L'] = -sqrt(2)*( C910Lmi * ha['+A+1-1'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['+T5+1-1'])
    A['Apara1', 'R'] = -sqrt(2)*( C910Rmi * ha['+A+1-1'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['+T5+1-1'])

    A['Aperp0', 'L'] = sqrt(2)*( C910Lpl * ha['0V+1+1'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['0T+1+1'])
    A['Aperp0', 'R'] = sqrt(2)*( C910Rpl * ha['0V+1+1'] - 2*mqh*(wc['7']+wc['7p'])/q2 * ha['0T+1+1'])
    A['Apara0', 'L'] = -sqrt(2)*( C910Lmi * ha['0A+1+1'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['0T5+1+1'])
    A['Apara0', 'R'] = -sqrt(2)*( C910Rmi * ha['0A+1+1'] + 2*mqh*(wc['7']-wc['7p'])/q2 * ha['0T5+1+1'])

    return {k: prefactor*v for k, v in A.items()}


def angular_coefficients(ta, br):
    # eqs (4.2) in arxiv
    # br is BR(L(1520)->pK)

    L={}
    L['1c'] = -2*br*( (ta['Aperp1','L'] * ta['Apara1','L'].conj()).real
                      - (ta['Aperp1','R'] * ta['Apara1','R'].conj()).real
    )
    L['1cc'] = br*( abs(ta['Apara1','L'])**2 + abs(ta['Aperp1','L'])**2
                    + abs(ta['Apara1','R'])**2 + abs(ta['Aperp1','R'])**2
    )
    L['1ss'] = br/2*( 2*(abs(ta['Apara0','L'])**2 + abs(ta['Aperp0','L'])**2)
                      + abs(ta['Apara1','L'])**2 + abs(ta['Aperp1','L'])**2
                      + 2*(abs(ta['Apara0','R'])**2 + abs(ta['Aperp0','R'])**2)
                      + abs(ta['Apara1','R'])**2 + abs(ta['Aperp1','R'])**2
    )
    L['2c'] = -br/2*( (ta['Aperp1','L'] * ta['Apara1','L'].conj()).real
                      + 3*(ta['Bperp1','L'] * ta['Bpara1','L'].conj()).real
                      - (ta['Aperp1','R'] * ta['Apara1','R'].conj()).real
                      - 3*(ta['Bperp1','R'] * ta['Bpara1','R'].conj()).real
    )
    L['2cc'] = br/4*( abs(ta['Apara1','L'])**2 + abs(ta['Aperp1','L'])**2
                      + 3*(abs(ta['Bpara1','L'])**2 + abs(ta['Bperp1','L'])**2)
                      + abs(ta['Apara1','R'])**2 + abs(ta['Aperp1','R'])**2
                      + 3*(abs(ta['Bpara1','R'])**2 + abs(ta['Bperp1','R'])**2)
    )
    L['2ss'] = br/8*( 2*abs(ta['Apara0','L'])**2 + abs(ta['Apara1','L'])**2
                      + 2*abs(ta['Aperp0','L'])**2 + abs(ta['Aperp1','L'])**2
                      + 3*(abs(ta['Bpara1','L'])**2 + abs(ta['Bperp1','L'])**2)
                      - 2*sqrt(3)*(ta['Bpara1','L']*ta['Apara1','L'].conj()).real
                      + 2*sqrt(3)*(ta['Bperp1','L']*ta['Aperp1','L'].conj()).real
                      + 2*abs(ta['Apara0','R'])**2 + abs(ta['Apara1','R'])**2
                      + 2*abs(ta['Aperp0','R'])**2 + abs(ta['Aperp1','R'])**2
                      + 3*(abs(ta['Bpara1','R'])**2 + abs(ta['Bperp1','R'])**2)
                      - 2*sqrt(3)*(ta['Bpara1','R']*ta['Apara1','R'].conj()).real
                      + 2*sqrt(3)*(ta['Bperp1','R']*ta['Aperp1','R'].conj()).real
    )
    L['3ss'] = sqrt(3)/2*br*( (ta['Bpara1','L']*ta['Apara1','L'].conj()).real
                              - (ta['Bperp1','L']*ta['Aperp1','L'].conj()).real
                              + (ta['Bpara1','R']*ta['Apara1','R'].conj()).real
                              - (ta['Bperp1','R']*ta['Aperp1','R'].conj()).real
    )
    L['4ss'] = sqrt(3)/2*br*( (ta['Bperp1','L']*ta['Apara1','L'].conj()).imag
                              - (ta['Bpara1','L']*ta['Aperp1','L'].conj()).imag
                              + (ta['Bperp1','R']*ta['Apara1','R'].conj()).imag
                              - (ta['Bpara1','R']*ta['Aperp1','R'].conj()).imag
    )
    L['5s'] = sqrt(3/2)*br*( (ta['Bperp1','L']*ta['Apara0','L'].conj()).real
                             - (ta['Bpara1','L']*ta['Aperp0','L'].conj()).real
                             - (ta['Bperp1','R']*ta['Apara0','R'].conj()).real
                             + (ta['Bpara1','R']*ta['Aperp0','R'].conj()).real
    )
    L['5sc'] = sqrt(3/2)*br*( -(ta['Bpara1','L']*ta['Apara0','L'].conj()).real
                              + (ta['Bperp1','L']*ta['Aperp0','L'].conj()).real
                              - (ta['Bpara1','R']*ta['Apara0','R'].conj()).real
                              + (ta['Bperp1','R']*ta['Aperp0','R'].conj()).real
    )
    L['6s'] = sqrt(3/2)*br*( (ta['Bpara1','L']*ta['Apara0','L'].conj()).imag
                             - (ta['Bperp1','L']*ta['Aperp0','L'].conj()).imag
                             - (ta['Bpara1','R']*ta['Apara0','R'].conj()).imag
                             + (ta['Bperp1','R']*ta['Aperp0','R'].conj()).imag
    )
    L['6sc'] = -sqrt(3/2)*br*( (ta['Bperp1','L']*ta['Apara0','L'].conj()).imag
                             - (ta['Bpara1','L']*ta['Aperp0','L'].conj()).imag
                             + (ta['Bperp1','R']*ta['Apara0','R'].conj()).imag
                             - (ta['Bpara1','R']*ta['Aperp0','R'].conj()).imag
    )

    return L


def prefactor(q2, par, scale):
    # calculate prefactor N
    xi_t = flavio.physics.ckm.xi('t','bs')(par)
    alphaem = flavio.physics.running.running.get_alpha(par, scale)['alpha_e']
    mLb = par['m_Lambdab']
    mL = par['m_Lambda(1520)']
    la_k = flavio.physics.bdecays.common.lambda_K(mLb**2, mL**2, q2)
    return par['GF'] * xi_t * alphaem * sqrt(q2) * la_k**(1/4.) / sqrt(3 * 2 * mLb**3 * pi**5) / 32


def get_ff(q2, par):
    ff_aux = AuxiliaryQuantity['Lambdab->Lambda(1520) form factor']
    return ff_aux.prediction(par_dict=par, wc_obj=None, q2=q2)


def get_transversity_amps_ff(q2, wc_obj, par_dict, lep, cp_conjugate):
    par = par_dict.copy()
    if cp_conjugate:
        par = conjugate_par(par)
    scale = flavio.config['renormalization scale']['lambdab']
    mLb = par['m_Lambdab']
    mL = par['m_Lambda(1520)']
    mb = flavio.physics.running.running.get_mb(par, scale)
    ff = get_ff(q2, par)
    wc = flavio.physics.bdecays.wilsoncoefficients.wctot_dict(wc_obj, 'bs' + lep + lep, scale, par)
    wc_eff = flavio.physics.bdecays.wilsoncoefficients.get_wceff(q2, wc, par, 'Lambdab', 'Lambda(1520)', lep, scale)

    ha = helicity_amps(q2, mLb, mL, ff)
    N = prefactor(q2, par, scale)
    ta_ff = transversity_amps(ha, q2, mLb, mL, mb, wc_eff, N, 'bs'+lep+lep)
    return ta_ff


def get_transversity_amps(q2, wc_obj, par, lep, cp_conjugate):
    if q2 >= 8.7 and q2 < 14:
        warnings.warn("The prediction in the region of narrow charmonium resonances are not meaningful")
    return get_transversity_amps_ff(q2, wc_obj, par, lep, cp_conjugate)


def get_obs(function, q2, wc_obj, par, lep, *ang_coef):
    ml = par['m_'+lep]
    mLb = par['m_Lambdab']
    mL = par['m_Lambda(1520)']
    if q2 < 4*ml**2 or q2 > (mLb-mL)**2:
        return 0
    ta = get_transversity_amps(q2, wc_obj, par, lep, cp_conjugate=False)

    br = par['BR(Lambda(1520)->NKbar)_exp']/2
    L = angular_coefficients(ta, br)
    ta_conj = get_transversity_amps(q2, wc_obj, par, lep, cp_conjugate=True)
    L_conj = angular_coefficients(ta_conj, br)
    return function(L, L_conj, *ang_coef)


# OBSERVABLES
def dGdq2(L, L_conj, *args):
    # differential decay width : (dG + dGbar)/dq2*(1/2)
    return (L['1cc'] + 2*L['1ss'] + 2*L['2cc'] + 4*L['2ss'] + 2*L['3ss'] + L_conj['1cc'] + 2*L_conj['1ss'] + 2*L_conj['2cc'] + 4*L_conj['2ss'] + 2*L_conj['3ss'])/6


def S(L, L_conj, ang_coef):
    # CP-averaged angular observalbes
    # ang_coef is for example '1cc'
    return ( L[ang_coef] + L_conj[ang_coef] )/2


def A(L, L_conj, ang_coef):
    # CP-asymmetries
    # ang_coef is for example '1cc'
    return ( L[ang_coef] - L_conj[ang_coef] )/2


def FL_num(L, L_conj, *args):
    # longuitudinal polarization of the dilepton system
    return (-L['1cc'] + 2*L['1ss'] - 2*L['2cc'] + 4*L['2ss'] + 2*L['3ss'] - L_conj['1cc'] + 2*L_conj['1ss'] - 2*L_conj['2cc'] + 4*L_conj['2ss'] + 2*L_conj['3ss'])/6


def AFBl_num(L, L_conj, *args):
    return (L['1c'] + 2*L['2c'] + L_conj['1c'] + 2*L_conj['2c'])/4


def AFBh(*args):
    return 0


def AFBlh(*args):
    return 0


def obs_int(function, q2min, q2max, wc_obj, par, lep, *ang_coef):
    def obs(q2):
        return get_obs(function, q2, wc_obj, par, lep, *ang_coef)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)


# Functions returning functions needed for Prediction instance

def dbrdq2_int_func(lep):
    def fct(wc_obj, par, q2min, q2max):
        tauLb = par['tau_Lambdab']
        return tauLb*obs_int(dGdq2, q2min, q2max, wc_obj, par, lep)/(q2max-q2min)
    return fct


def dbrdq2_func(lep):
    def fct(wc_obj, par, q2):
        tauLb = par['tau_Lambdab']
        return tauLb * get_obs(dGdq2, q2, wc_obj, par, lep)
    return fct


def obs_ratio_func(func_num, func_den, lep, *ang_coef):
    def fct(wc_obj, par, q2):
        num = get_obs(func_num, q2, wc_obj, par, lep, *ang_coef)
        if num == 0:
            return 0
        denom = get_obs(func_den, q2, wc_obj, par, lep, *ang_coef)
        return num/denom
    return fct


def obs_int_ratio_func(func_num, func_den, lep, *ang_coef):
    def fct(wc_obj, par, q2min, q2max):
        num = obs_int(func_num, q2min, q2max, wc_obj, par, lep, *ang_coef)
        if num == 0:
            return 0
        denom = obs_int(func_den, q2min, q2max, wc_obj, par, lep, *ang_coef)
        return num/denom
    return fct


_tex = {'e': 'e', 'mu': r'\mu', 'tau': r'\tau'}
_observables = {
    'FL': {'func_num': FL_num, 'tex': r'F_L', 'desc': 'longitudinal polarization fraction'},
    'AFBl': {'func_num': AFBl_num, 'tex': r'A_\text{FB}^\ell', 'desc': 'leptonic forward-backward asymmetry'},
    'AFBh': {'func_num': AFBh, 'tex': r'A_\text{FB}^\ell', 'desc': 'hadronic forward-backward asymmetry'},
    'AFBlh': {'func_num': AFBlh, 'tex': r'A_\text{FB}^{\ell h}', 'desc': 'lepton-hadron forward-backward asymmetry'}
    }

# subscript of angular coefficients L
ang_coef_List = ['1c', '1cc', '1ss', '2c', '2cc', '2ss', '3ss', '4ss', '5s', '5sc', '6s', '6sc']

for a in ang_coef_List:
    Sstring = 'S_'+a
    Astring = 'A_'+a
    _observables[Sstring] = {'func_num': S, 'tex': r'S_{'+a+'}', 'desc': 'CP symmetry '+a, 'ang_coef': a}
    _observables[Astring] = {'func_num': A, 'tex': r'A_{'+a+'}', 'desc': 'CP asymmetry '+a, 'ang_coef': a}


for l in ['e', 'mu', ]:

    _process_tex = r'\Lambda_b\to\Lambda(1520) '+_tex[l]+r'^+'+_tex[l]+r'^-'
    _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $\Lambda_b\to\Lambda(1520)\ell^+\ell^-$ :: $' + _process_tex + r'$'

    # binned branching ratio
    _obs_name = "<dBR/dq2>(Lambdab->Lambda(1520)"+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
    _obs.set_description(r"Binned differential branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\langle \frac{d\text{BR}}{dq^2} \rangle(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    Prediction(_obs_name, dbrdq2_int_func(l))

    # differential branching ratio
    _obs_name = "dBR/dq2(Lambdab->Lambda(1520)"+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2'])
    _obs.set_description(r"Differential branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\frac{d\text{BR}}{dq^2}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    Prediction(_obs_name, dbrdq2_func(l))

    for obs in _observables:
        # binned angular observables
        _obs_name = "<" + obs+">(Lambdab->Lambda(1520)"+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description("Binned " + _observables[obs]['desc'] + r" in $" + _process_tex + r"$")
        _obs.tex = r"$\langle " + _observables[obs]['tex'] + r"\rangle(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        if 'ang_coef' in _observables[obs].keys():
            Prediction(_obs_name, obs_int_ratio_func(_observables[obs]['func_num'], dGdq2, l, _observables[obs]['ang_coef']))
        else :
            Prediction(_obs_name, obs_int_ratio_func(_observables[obs]['func_num'], dGdq2, l))

        # differential angular observables
        _obs_name = obs+"(Lambdab->Lambda(1520)"+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(_observables[obs]['desc'][0].capitalize() + _observables[obs]['desc'][1:] + r" in $" + _process_tex + r"$")
        _obs.tex = r"$" + _observables[obs]['tex'] + r"(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        if 'ang_coef' in _observables[obs].keys():
            Prediction(_obs_name, obs_ratio_func(_observables[obs]['func_num'], dGdq2, l, _observables[obs]['ang_coef']))
        else :
            Prediction(_obs_name, obs_ratio_func(_observables[obs]['func_num'], dGdq2, l))



import unittest
from math import sqrt,radians,asin
from flavio.physics.bdecays.formfactors.b_p import btop, bcl_parameters, bcl, bsz_parameters
from flavio.physics.bdecays.formfactors.b_v.test_btov import test_eos_ff
import numpy as np
import copy
from flavio.parameters import default_parameters
from flavio.classes import Parameter, Implementation

c = copy.deepcopy(default_parameters)

class TestBtoP(unittest.TestCase):
    def test_bcl_iw(self):
        c = copy.deepcopy(default_parameters)
        par = c.get_central_all()
        bcl.ff_isgurwise('B->D', 1, par, 4.8, n=3)
        Implementation['B->D BCL3-IW'].get_central(constraints_obj=c, wc_obj=None, q2=1)
        # assert f+(0)=f0(0)
        ffq20 = Implementation['B->D BCL3-IW'].get_central(constraints_obj=c, wc_obj=None, q2=0)
        self.assertEqual(ffq20['f+'], ffq20['f0'])

    def test_lattice(self):
        # compare to digitized fig. 18 of 1509.06235v1
        ff_latt = Implementation['B->K BCL3'].get_central(constraints_obj=c, wc_obj=None, q2=0)
        self.assertAlmostEqual(ff_latt['f+'], 0.33, places=1)
        self.assertAlmostEqual(ff_latt['f0'], 0.33, places=1)
        self.assertAlmostEqual(ff_latt['fT'], 0.26, places=1)
        ff_latt = Implementation['B->K BCL3'].get_central(constraints_obj=c, wc_obj=None, q2=5)
        self.assertAlmostEqual(ff_latt['f+'], 0.42, places=1)
        self.assertAlmostEqual(ff_latt['f0'], 0.36, places=1)
        self.assertAlmostEqual(ff_latt['fT'], 0.37, places=1)
        ff_latt = Implementation['B->K BCL3'].get_central(constraints_obj=c, wc_obj=None, q2=10)
        self.assertAlmostEqual(ff_latt['f+'], 0.64, places=1)
        self.assertAlmostEqual(ff_latt['f0'], 0.44, places=1)
        self.assertAlmostEqual(ff_latt['fT'], 0.57, places=1)
        ff_latt = Implementation['B->K BCL3'].get_central(constraints_obj=c, wc_obj=None, q2=15)
        self.assertAlmostEqual(ff_latt['f+'], 0.98, places=1)
        self.assertAlmostEqual(ff_latt['f0'], 0.56, places=1)
        self.assertAlmostEqual(ff_latt['fT'], 0.96, places=1)
        ff_latt = Implementation['B->K BCL3'].get_central(constraints_obj=c, wc_obj=None, q2=20)
        self.assertAlmostEqual(ff_latt['f+'], 1.7, places=1)
        self.assertAlmostEqual(ff_latt['f0'], 0.71, places=1)
        self.assertAlmostEqual(ff_latt['fT'], 1.74, places=1)

    def test_bsz(self):
        bsz_parameters.load_parameters('data/arXiv-1811-00983v1/BD_LCSR.json', 'B->D', c)
        ff_latt = Implementation['B->D BSZ3'].get_central(constraints_obj=c, wc_obj=None, q2=0)

    def test_gkvd(self):
        # compare to numbers of arXiv:1811.00983
        c = copy.deepcopy(default_parameters)
        for q2 in [1.5, 6]:
            for ff in ['f+', 'f0']:
                for P in ['K', 'D', 'pi']:
                    bsz_parameters.gkvd_load('v1', 'LCSR-Lattice', ('B->{}'.format(P),), c)
                    ffbsz3 = Implementation['B->{} BSZ3'.format(P)].get_central(constraints_obj=c, wc_obj=None, q2=q2)
                    self.assertAlmostEqual(ffbsz3[ff] / test_eos_ff[P][q2][ff],
                                           1,
                                           places=3,
                                           msg="Failed for {} in B->{} at q2={}".format(ff, P, q2))

    def test_bcl_lmvd(self):
        # compare to results obtained from EOS (see https://gist.github.com/peterstangl/7d6c862bff87a10e7334993acd2ae0c5 for the notebook used)
        # see also fig. 5 in arXiv:2102.07233
        q2vals = [-10, -5, 0, 5, 10, 15, 20, 25]
        eos_Btopi = np.array([[0.13944073, 0.1786386 , 0.23467602, 0.31969463,
                               0.46026432, 0.7264543 , 1.37263015, 4.34113315],
                              [0.19729307, 0.21337439, 0.23467602, 0.26452324,
                               0.30960826, 0.38517962, 0.53237519, 0.87948273],
                              [0.14937763, 0.18405398, 0.23500764, 0.31456728,
                               0.44987482, 0.71224394, 1.35505172, 4.19770336]])

        for i, q2 in enumerate(q2vals):
            ff_bcl_lmvd = Implementation['B->pi BCL4-LMVD'].get_central(constraints_obj=c, wc_obj=None, q2=q2)
            self.assertAlmostEqual(ff_bcl_lmvd['f+'], eos_Btopi[0,i], places=5, msg="Failed for f+ in B->pi at q2={}".format(q2))
            self.assertAlmostEqual(ff_bcl_lmvd['f0'], eos_Btopi[1,i], places=5, msg="Failed for f0 in B->pi at q2={}".format(q2))
            self.assertAlmostEqual(ff_bcl_lmvd['fT'], eos_Btopi[2,i], places=5, msg="Failed for fT in B->pi at q2={}".format(q2))



import unittest
import numpy as np
import flavio

constraints = flavio.default_parameters
wc_obj = flavio.WilsonCoefficients()
par = constraints.get_central_all()

class TestKpilnu(unittest.TestCase):
    def test_kpilnu(self):
        # test for errors
        q2=0.05
        flavio.physics.kdecays.kpilnu.get_ff(q2, par, 'KL')
        flavio.physics.kdecays.kpilnu.get_ff(q2, par, 'K+')
        flavio.physics.kdecays.kpilnu.get_angularcoeff(q2, wc_obj, par, 'KL', 'pi+', 'e')
        flavio.physics.kdecays.kpilnu.get_angularcoeff(q2, wc_obj, par, 'KL', 'pi+', 'mu')
        flavio.physics.kdecays.kpilnu.get_angularcoeff(q2, wc_obj, par, 'K+', 'pi0', 'e')
        flavio.physics.kdecays.kpilnu.get_angularcoeff(q2, wc_obj, par, 'K+', 'pi0', 'mu')
        # unphysical q2
        self.assertEqual(flavio.physics.kdecays.kpilnu.dBRdq2(0, wc_obj, par, 'KL', 'pi+', 'e'), 0)
        self.assertEqual(flavio.physics.kdecays.kpilnu.dBRdq2(0.01, wc_obj, par, 'K+', 'pi0', 'mu'), 0)
        self.assertEqual(flavio.physics.kdecays.kpilnu.dBRdq2(2, wc_obj, par, 'K+', 'pi0', 'e'), 0)
        self.assertEqual(flavio.physics.kdecays.kpilnu.dBRdq2(1.5, wc_obj, par, 'KL', 'pi+', 'mu'), 0)
        self.assertEqual(flavio.physics.kdecays.kpilnu.dBRdq2(0.1, wc_obj, par, 'KL', 'pi+', 'tau'), 0)
        # compare central predictions to PDG values
        self.assertAlmostEqual(flavio.sm_prediction('BR(KL->pienu)')*1e2/40.55, 1, delta=0.04)
        self.assertAlmostEqual(flavio.sm_prediction('BR(K+->pienu)')*1e2/5.07, 1, delta=0.04)
        self.assertAlmostEqual(flavio.sm_prediction('BR(KL->pimunu)')*1e2/27.04, 1, delta=0.02)
        self.assertAlmostEqual(flavio.sm_prediction('BR(K+->pimunu)')*1e2/3.352, 1, delta=0.03)


    def test_kpilnu_nu(self):
        wc_sm = flavio.WilsonCoefficients()
        wc_np_mu = flavio.WilsonCoefficients()
        wc_np_mu.set_initial({'CVL_sumunumu': 1}, 4.8)
        wc_np_e = flavio.WilsonCoefficients()
        wc_np_e.set_initial({'CVL_sumunue': 1}, 4.8)
        obs = flavio.Observable["BR(K+->pimunu)"]
        br_sm = obs.prediction_central(constraints, wc_sm)
        br_mu = obs.prediction_central(constraints, wc_np_mu)
        br_e = obs.prediction_central(constraints, wc_np_e)
        # with interference: (1 + 1)^2 = 4
        self.assertAlmostEqual(br_mu/br_sm, 4, delta=0.06)
        # without interference: 1 + 1 = 2
        self.assertAlmostEqual(br_e/br_sm, 2, delta=0.03)

    def test_logC(self):
        # SM: prediction = input
        self.assertAlmostEqual(flavio.sm_prediction('lnC(K->pimunu)'), par['K->pi ln(C)'], delta=0.0001)

    def test_RT(self):
        self.assertEqual(flavio.sm_prediction('RT(K->pimunu)'), 0)
        wc = flavio.WilsonCoefficients()
        wc.set_initial({'CT_sumunumu': 1}, par['m_rho0'])
        self.assertEqual(flavio.np_prediction('RT(K->pienu)', wc), 0)
        self.assertNotEqual(flavio.np_prediction('RT(K->pimunu)', wc), 0)



from math import sqrt, exp
import flavio

def lambda_momentum(q2, mL, mLb):
    # daughter baryon momentum in the Lb rest frame
    s = q2/(mLb**2)
    r = (mL/mLb)**2
    phi = (1-r)**2 - 2*(1+r)*s + s**2
    if phi > 0 :
        return mLb/2*sqrt(phi)
    else :
        return mLb/2*sqrt(-phi)

def lambda_mass(m_q, m_s):
    return 2*m_q + m_s


def alpha_lambda_lambdaprime(alpha_l1, alpha_l2):
    return sqrt((alpha_l1**2 + alpha_l2**2)/2)


def F(a0, a2, a4, p_L, m_q, m_L, alpha):
    return (a0 + a2*p_L**2 + a4*p_L**4)*exp(-3*(m_q*p_L)**2/(2*(m_L*alpha)**2))


_process_dict = {}
_process_dict['Lambdab->Lambda(1520)'] = {'X': 'Lambda(1520)'}


def formfactors(process, par, q2):
    r"Functions for $\Lambda_b\to X_{3/2}$ form factors where $X_{3/2} is a spin-3/2 baryon$ using the Quark Model and the MCN approach treated in arXiv:1108.6129 [nucl-th]"
    flavio.citations.register("Mott:2011cx")

    # Using the PDG mass values instead the model ones, will be covered by the uncertainties attached to the form factors.
    pd = _process_dict[process]
    mL = par['m_Lambda(1520)']
    mLb = par['m_Lambdab']
    m_q = par[process + ' m_q']
    m_s = par[process + ' m_s']
    alpha_l1 = par[process +' alpha_Lambdab']
    alpha_l2 = par[process +' alpha_'+pd['X']]

    ff_list = ['F1', 'F2', 'F3', 'F4',
               'G1', 'G2', 'G3', 'G4',
               'H1', 'H2', 'H3', 'H4', 'H5', 'H6']

    p_lambda = lambda_momentum(q2, mL, mLb)
    m_lambda_tilde = lambda_mass(m_q, m_s)
    alpha_ll = alpha_lambda_lambdaprime(alpha_l1, alpha_l2)

    ff_dict = {}
    for e in ff_list:
        a0 = par[process+' '+e+' a0']
        a2 = par[process+' '+e+' a2']
        a4 = par[process+' '+e+' a4']

        ff_dict[e] = F(a0, a2, a4, p_lambda, m_q, m_lambda_tilde, alpha_ll)

    return ff_dict, mL, mLb


def ff_equiv(process, q2, par):
    # transform FormDict in form factors used in arXiv:1903.00448
    ff_dict, mL, mLb = formfactors(process, par, q2)

    e_fVt = par[process+' fVt uncertainty']
    e_fVperp = par[process+' fVperp uncertainty']
    e_fV0 = par[process+' fV0 uncertainty']
    e_fVg = par[process+' fVg uncertainty']
    e_fAt = par[process+' fAt uncertainty']
    e_fAperp = par[process+' fAperp uncertainty']
    e_fA0 = par[process+' fA0 uncertainty']
    e_fAg = par[process+' fAg uncertainty']
    e_fTt = par[process+' fTt uncertainty']
    e_fTperp = par[process+' fTperp uncertainty']
    e_fT0 = par[process+' fT0 uncertainty']
    e_fTg = par[process+' fTg uncertainty']
    e_fT5t = par[process+' fT5t uncertainty']
    e_fT5perp = par[process+' fT5perp uncertainty']
    e_fT50 = par[process+' fT50 uncertainty']
    e_fT5g = par[process+' fT5g uncertainty']

    ff = {}
    ff['fVt'] = ( ff_dict['F2']*mL*(mL**2 - mLb**2 - q2) + mLb*(2*ff_dict['F1']*mL*(mL - mLb) - 2*ff_dict['F4']*mL*mLb + ff_dict['F3']*(mL**2 - mLb**2 + q2)) )/( 2*mL*(mL-mLb)*mLb**2 ) * e_fVt
    ff['fVperp'] = ( ff_dict["F1"]/mLb - ff_dict["F4"]*mL/(mL**2 - 2*mL*mLb + mLb**2 - q2) )*e_fVperp
    ff['fV0'] = ( ff_dict["F2"]*mL*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2)) + mLb*(2*ff_dict["F1"]*mL*(mL + mLb)*(mL**2 - 2*mL*mLb + mLb**2 - q2) - 2*ff_dict["F4"]*mL*mLb*(mL**2 - mLb**2 + q2) + ff_dict["F3"]*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2))) )/( 2*mL*mLb**2*(mL+mLb)*(mL**2 - 2*mL*mLb + mLb**2 - q2) )*e_fV0
    ff['fVg'] = ff_dict["F4"]*e_fVg

    ff['fAt'] = ( ff_dict["G2"]*mL*(mL**2 - mLb**2 - q2) + mLb*(-2*ff_dict["G4"]*mL*mLb + 2*ff_dict["G1"]*mL*(mL + mLb) + ff_dict["G3"]*(mL**2 - mLb**2 + q2)) )/( 2*mL*mLb**2*(mL + mLb) )*e_fAt
    ff['fAperp'] = ( ff_dict["G1"]/mLb - (ff_dict["G4"]*mL)/(mL**2 + 2*mL*mLb + mLb**2 - q2) )*e_fAperp
    ff['fA0'] = ( ff_dict["G2"]*mL*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2)) + mLb*(2*ff_dict["G1"]*mL*(mL - mLb)*(mL**2 + 2*mL*mLb + mLb**2 - q2) - 2*ff_dict["G4"]*mL*mLb*(mL**2 - mLb**2 + q2) + ff_dict["G3"]*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2))) )/( 2*mL*(mL - mLb)*mLb**2*(mL**2 + 2*mL*mLb + mLb**2 - q2) )*e_fA0
    ff['fAg'] = -ff_dict["G4"]*e_fAg

    ff['fTt'] = 0*e_fTt
    ff['fTperp'] = ( 2*ff_dict["H5"]*mL - ((ff_dict["H3"]+ff_dict["H6"])*mL**2)/mLb + ff_dict["H3"]*mLb + 2*ff_dict["H1"]*mL*(mL + mLb)/mLb - 2*(ff_dict["H5"] + ff_dict["H6"])*mL**2*(mL - mLb)/((mL - mLb)**2 - q2) - ff_dict["H3"]*q2/mLb + ff_dict["H2"]*mL*(-mL**2 + mLb**2 + q2)/mLb**2 )/( 2*mL*(mL + mLb) )*e_fTperp
    ff['fT0'] = ( (ff_dict["H1"] + ff_dict["H2"] - ff_dict["H3"] - ff_dict["H6"])/mLb - 2*((ff_dict["H5"] + ff_dict["H6"])*mL)/((mL - mLb)**2 - q2) + ff_dict["H4"]*((mL + mLb)**2 - q2)/(2*mL*mLb**2) )*e_fT0
    ff['fTg'] = ( ff_dict["H5"]*(mL- mLb) - ff_dict["H6"]*(-mL**2 + mLb**2 + q2)/(2*mLb) )*e_fTg

    ff['fT5t'] = 0*e_fT5t
    ff['fT5perp'] = ( -1/(2*mL*(mL-mLb)*mLb**2*(mL**2 + 2*mL*mLb + mLb**2 - q2)) * (ff_dict["H2"]*mL*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2)) + mLb*(mL*(2*ff_dict["H5"]*mLb*(mL*mLb + mLb**2 - q2) + ff_dict["H6"]*mL*(mL**2 + 2*mL*mLb + mLb**2 -q2)) - 2*ff_dict["H1"]*mL*(mL - mLb)*(mL**2 + 2*mL*mLb + mLb**2 -q2) + ff_dict["H3"]*(mL**4 + (mLb**2 - q2)**2 - 2*mL**2*(mLb**2 + q2)))) )*e_fT5perp
    ff['fT50'] = ( ff_dict["H1"]/mLb + 2*ff_dict["H5"]*mL/(mL**2 + 2*mL*mLb + mLb**2 - q2) )*e_fT50
    ff['fT5g'] = ( -ff_dict["H5"]*(mL + mLb) - ff_dict["H6"]*(mL**2 + 2*mL*mLb + mLb**2 - q2)/(2*mLb) )*e_fT5g

    return ff



from flavio.math.integrate import nintegrate
from flavio.physics.common import add_dict
from flavio.physics import ckm as ckm_flavio
from flavio.classes import Observable, Prediction
from . import partondist
import wilson
from numpy import pi, sqrt
import numpy as np
from flavio.config import config

def F_qqlnu_SM(s, par):
    r"""flavour-independent SM $W-$ contribution to the $\bar u d\to \ell^- \nu_\ell$ amplitude."""
    # parameters
    mW = par['m_W']
    GammaW = 1 / par['tau_W']
    s2w = par['s2w']
    aEW = par['alpha_e']
    e2 = 4 * pi * aEW
    g2=e2/s2w * 0.9608 # correction factor from running g^2 to 1TeV

    return g2/2/(s-mW**2 + 1j*mW*GammaW)


def wceff_qqlnu_sm(s, par):
    ckm = ckm_flavio.get_ckm(par)
    E = np.einsum('ij,kl->ijkl', np.eye(3), np.eye(3))
    wc = {}
    # include CKM matrix here
    wc['CVLL_enuud'] = np.einsum('ijml,km->ijkl',F_qqlnu_SM(s, par)*E,ckm)
    wc['CSRL_enuud'] = wc['CSRR_enuud'] = wc['CTRR_enuud'] = np.zeros((3, 3, 3, 3))

    return wc

def wceff_qqlnu_np(wc_obj, par, scale):
    r"""Returns Wilson coefficients of the effective Lagrangian
    $$\mathcal{L} = C_{e \nu u d}^{\Gamma_1\Gamma_2}
    (\bar e_i\Gamma_1 \nu_j)(\bar u_k\Gamma_2 d_l)$$
    as a dictionary of arrays with shape (3, 3, 3, 3) corresponding to ijkl.
    """
    # get the dictionary
    wcxf_dict = wc_obj.get_wcxf(sector='dB=dL=0', scale=scale, par=par,
                                eft='SMEFT', basis='Warsaw up').dict
    # go to redundant basis, C has all the smeft coefficients
    C = wilson.util.smeftutil.wcxf2arrays_symmetrized(wcxf_dict)

    wc = {}
    ckm = ckm_flavio.get_ckm(par)
    # match to notation used in the observable
    wc['CVLL_enuud'] = 2*np.einsum('ijkm,ml->ijkl',C['lq3'],ckm)
    wc['CSRL_enuud'] = np.einsum('jilk',np.conjugate(C['ledq']))
    wc['CSRR_enuud'] = np.einsum('jimk,ml->ijkl',np.conjugate(C['lequ1']),ckm)
    wc['CTRR_enuud'] = np.einsum('jimk,ml->ijkl',np.conjugate(C['lequ3']),ckm)

    return wc

fermion_indices = {'u': 0, 'd': 0, 'c': 1, 's': 1, 't': 2, 'b': 2, 'e': 0, 'mu': 1, 'tau': 2}

def dsigma_dtau_qqlnu_partial(s, tau, umin, umax, q1, q2, l, wc_eff):
    r"""Partial result for the cross section of $\bar q1 q2 \to \ell^-\nu$.

    Returns $\int_{umin}^{umax} \frac{d^2\sigma}{d\tau du}$, integrated analytically over $u=m_T^2/s$.

    Parameters:
    - `s`: Hadronic center of mass energy GeV$^2$
    - `tau`: $\tau$, ratio of partonic to hadronic center of mass energy, dimensionless
    - `umin`, `umax: integration boundaries for the variable $u=m_T^2/s$
    - `q1, q2`: initial state quarks (with q1 up-type)
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `wc_eff`: SM + SMEFT amplitude for the process as dictionary with entries corresponding to different Wilson coefficients
    """
    i1 = fermion_indices[q1]
    i2 = fermion_indices[q2]
    il = fermion_indices[l]

    if tau > umax:
        prefac1 = umax * sqrt(1-umax/tau) - umin * sqrt(1-umin/tau)
        prefac2 = tau * ( sqrt(1-umax/tau) - sqrt(1-umin/tau) )
    else:
        prefac1 = - umin * sqrt(1-umin/tau)
        prefac2 = - tau * sqrt(1-umin/tau)

    S = 0
    # sum over neutrino flavours
    for inu in range(3):
        S += s * (
            prefac1 * (
                + 16 * abs(wc_eff['CTRR_enuud'][il, inu, i1, i2])**2
                + abs(wc_eff['CVLL_enuud'][il, inu, i1, i2])**2
            )
            - prefac2 *(
                + 3 * abs(wc_eff['CSRL_enuud'][il, inu, i1, i2])**2
                + 3 * abs(wc_eff['CSRR_enuud'][il, inu, i1, i2])**2
                + 16 * abs(wc_eff['CTRR_enuud'][il, inu, i1, i2])**2
                + 4 * abs(wc_eff['CVLL_enuud'][il, inu, i1, i2])**2
            )
        )

    return S/(576*pi)

pdf_flavor = {'d': 1, 'u': 2, 's': 3, 'c': 4, 'b': 5}

def dsigma_dtau_qqlnu(s, tau, umin, umax, l, wc_eff, par, Q2):
    r"""Differential hadronic cross section of $pp\to \ell\nu$.

    Returns ${d\sigma}{d\tau}$ in units of GeV$^{-2}$.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `tau`: $\tau$, ratio of partonic to hadronic center of mass energy, dimensionless
    - `umin`, `umax`: integration boundaries for the variable $u=m_T^2/s$
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `wc_eff`: SM + SMEFT amplitude for the process as dictionary with entries corresponding to different Wilson coefficients
    - `par`: parameter dictionary
    - `Q2`: factorization scale squared in GeV$^2$
    """
    sigma = 0
    members_par = config['PDF set']['dileptons']['members par']
    member = int(par[members_par])
    plumi = partondist.get_parton_lumi(Q2=Q2, member=member)
    for q1 in 'uc':
        for q2 in 'dsb':
            i1 = pdf_flavor[q1]
            i2 = pdf_flavor[q2]
            y = dsigma_dtau_qqlnu_partial(s, tau, umin, umax, q1, q2, l, wc_eff)
            if y != 0:  # save time if the contribution is anyway zero
                sigma += 2 * plumi.L(i1, -i2, tau) * y # $W^+$ contribution
                sigma += 2 * plumi.L(i2, -i1, tau) * y # $W^-$ contribution
    return sigma

def sigma_qqlnu_int(s, mTmin, mTmax, l, wc_obj, par, Q2, scale, newphys=True):
    r"""Integrated hadronic cross section of $pp\to \ell\nu$.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `mTmin`, `mTmax`: bin boundaries in transverse mass mT in GeV
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `wc_obj`: Wilson coefficient object
    - `par`: parameter dictionary
    - `Q2`: factorization scale squared in GeV$^2$
    - `scale`: scale at which the Wilson coefficients are evaluated in GeV
    - `newphys`: boolean, whether to include NP contributions or not
    """
    # change of variables to $u=m_T^2/s$
    umin = mTmin**2/s
    umax = mTmax**2/s

    if newphys:
        wc_eff_np = wceff_qqlnu_np(wc_obj, par, scale)
    else:
        wc_eff_np = {}
    def f(tau):
        wc_eff = wceff_qqlnu_sm(tau * s, par)
        if newphys:
            # add NP contribution
            wc_eff = add_dict((wc_eff, wc_eff_np))
        return dsigma_dtau_qqlnu(s, tau, umin, umax, l, wc_eff, par,  Q2)
    # integrate over tau from umin to 1
    return nintegrate(f, umin, 1, epsrel=1e-5)

def R_sigma_qqlnu_int(s, mTmin, mTmax, l, wc_obj, par):
    r"""Integrated hadronic cross section of $pp\to \ell \nu$ normalized
    to its SM value.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `mTmin`, `mTmax`: bin boundaries in transverse mass mT in GeV
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `wc_obj`: Wilson coefficient object
    - `par`: parameter dictionary
    """
    # Renormalization and factorization scale are fixed in the bin
    scale = (mTmin + mTmax)/2
    Q2 = scale**2
    sigma_sm = sigma_qqlnu_int(s, mTmin, mTmax, l, wc_obj, par, Q2, scale, newphys=False)
    sigma_np = sigma_qqlnu_int(s, mTmin, mTmax, l, wc_obj, par, Q2, scale, newphys=True)
    return sigma_np / sigma_sm


def R_sigma_qqlnu_int_fct_lhc13(l):
    s = 13000**2
    def f(wc_obj, par, mTmin, mTmax):
        return R_sigma_qqlnu_int(s, mTmin, mTmax, l, wc_obj, par)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for l in _tex:
    _process_tex = r"pp\to " + _tex[l]+r"\nu"
    _process_taxonomy = r'Process :: contact interactions :: $' + _process_tex + r"$"

    _obs_name = "R_13(pp->" + l + "nu)"
    _obs = Observable(_obs_name, arguments=['mTmin', 'mTmax'])
    _obs.set_description(r"Cross section of $" + _process_tex + r"$ at $\sqrt{s}=13$ TeV normalized to the SM contribution")
    _obs.tex = r"$\text{R}_{13}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    Prediction(_obs_name, R_sigma_qqlnu_int_fct_lhc13(l))



r"""$V\to ll^\prime$ branching ratio"""

from flavio.classes import Observable, Prediction
from flavio.physics.running import running
import flavio
from flavio.physics.common import lambda_K
import numpy as np

meson_quark = { 'J/psi' : 'cc', 
                'psi(2S)': 'cc',
                'Upsilon(1S)': 'bb',
                'Upsilon(2S)': 'bb',
                'Upsilon(3S)': 'bb',
                }

wc_sector = { ('e','e') : "dF=0", ('mu','mu'): "dF=0", ('tau','tau'): "dF=0", ('e','mu') : 'mue',  ('mu','e') :'mue', ('e','tau') : 'taue', ('tau','e') : 'taue', ('mu','tau') : 'mutau', ('tau','mu') : 'mutau'}

def getVT_lfv(wc_obj,par,V,Q,l1,l2):
    # renormalization scale
    scale = flavio.config['renormalization scale'][V]
    # Wilson coefficients
    wc = wc_obj.get_wc(wc_sector[(l1,l2)], scale, par)

    alphaem = running.get_alpha_e(par, scale)
    ee=np.sqrt(4.*np.pi*alphaem) 

    mV = par['m_'+V]   
    fV=par['f_'+V] 
    fV_T=par['fT_'+V]
    qq=meson_quark[V]
    # for emu and taue the name of the Wilson coefficient sector agrees with the ordering of leptons in the vector bilinear
    # This is not the case for mutau. Thus distinguish between the two cases here.
    if wc_sector[(l1,l2)]=="mutau":
        ll="taumu"
    else:
        ll=wc_sector[(l1,l2)]
    VL=fV*mV*(wc['CVLL_'+ll+qq] + wc['CVLR_'+ll+qq]) 
    VR=fV*mV*(wc['CVRR_'+ll+qq] + wc['CVLR_'+qq+ll]) 
    TR=fV_T*mV*wc['CTRR_'+l1+l2+qq] - ee*Q *fV*wc['Cgamma_'+l2+l1] 
    TL=(fV_T*mV*wc['CTRR_'+l2+l1+qq] - ee*Q *fV*wc['Cgamma_'+l1+l2]).conjugate()
    if ll==l2+l1: 
        VL=VL.conjugate()
        VR=VR.conjugate()
    return VL,VR,TL,TR

def getVT_lfc(wc_obj,par,V,Q,l):
# add contribution from photon exchange and then activate LFC channels

    # renormalization scale
    scale = flavio.config['renormalization scale'][V]
    # Wilson coefficients
    wc = wc_obj.get_wc(wc_sector[(l,l)], scale, par)
    ll=l+l
    alphaem = running.get_alpha_e(par, scale)
    ee2=4.*np.pi*alphaem
    ee=np.sqrt(ee2) 

    norm=4*par['GF']/np.sqrt(2)
    normDipole=norm*ee/(16*np.pi**2)*par['m_'+l]

    mV = par['m_'+V]   
    fV=par['f_'+V] 
    fV_T=par['fT_'+V]
    qq=meson_quark[V]

    VL=fV*mV*(wc['CVLL_'+ll+qq]*norm + wc['CVLR_'+ll+qq]*norm - 2*Q* ee2/mV**2 ) 
    VR=fV*mV*(wc['CVRR_'+ll+qq]*norm + wc['CVLR_'+qq+ll]*norm  - 2*Q* ee2/mV**2 )
    TR=fV_T*mV*wc['CTRR_'+ll+qq]*norm - ee*Q *fV*wc['C7_'+ll]*normDipole 
    TL=TR.conjugate()
    
    return VL,VR,TL,TR


def Vll_br(wc_obj, par,V,Q, l1,l2):
    r"""Branching ratio for the lepton-flavour violating leptonic decay J/psi-> l l'"""
    #####branching ratio obtained from 2207.10913#####
    flavio.citations.register("Calibbi:2022ddo")
    mV = par['m_'+V]   
    tauV = par['tau_'+V]  
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    y1=ml1/mV
    y2=ml2/mV
    y1s=y1**2
    y2s=y2**2
    if l1==l2:
        VL,VR,TL,TR = getVT_lfc(wc_obj,par,V,Q,l1)
    else:
        VL,VR,TL,TR = getVT_lfv(wc_obj,par,V,Q,l1,l2)


    ampSquared_V = (np.abs(VL)**2+np.abs(VR)**2 )/12. *(2-y1s-y2s-(y1s-y2s)**2)+y1*y2*(VL*VR.conjugate()).real
    ampSquared_T= 4./3.*(np.abs(TL)**2+np.abs(TR)**2) * (1+y1s+y2s-2*(y1s-y2s)**2) +16.*y1*y2*(TR*TL.conjugate()).real
    ampSquared_VT = 2*y1*(1+y2s-y1s)*(VR*TR.conjugate()+VL*TL.conjugate()).real + 2*y2*(1+y1s-y2s)*(VL*TR.conjugate()+VR*TL.conjugate()).real
    
    return tauV*mV/(16.*np.pi) * np.sqrt(lambda_K(1,y1s,y2s)) * (ampSquared_V+ampSquared_T+ampSquared_VT)


def Vll_br_func(V, Q, l1, l2):
    def fct(wc_obj, par):
        return Vll_br(wc_obj, par, V, Q, l1, l2)
    return fct


def Vll_br_comb_func(V, Q, l1, l2):
    def fct(wc_obj, par):
        return Vll_br(wc_obj, par, V, Q, l1, l2)+ Vll_br(wc_obj, par, V, Q, l2, l1)
    return fct

def Vll_ratio_func(V, Q, l1, l2):
    def fct(wc_obj, par):
        BRee=Vll_br(wc_obj,par,V,Q,'e','e')
        return Vll_br(wc_obj, par, V, Q, l1, l2)/BRee
    return fct


def Vll_ratio_comb_func(V, Q, l1, l2):
    def fct(wc_obj, par):
        BRee=Vll_br(wc_obj,par,V,Q,'e','e')
        return (Vll_br(wc_obj, par, V, Q, l1, l2)+ Vll_br(wc_obj, par, V, Q, l2, l1))/BRee
    return fct


# Observable and Prediction instances
_hadr = { 
    'J/psi': {'tex': r"J/\psi\to", 'V': 'J/psi', 'Q': 2./3., },
    'psi(2S)': {'tex': r"\psi(2S)\to", 'V': 'psi(2S)', 'Q': 2./3., }, 
    'Upsilon(1S)': {'tex': r"\Upsilon(1S)\to", 'V': 'Upsilon(1S)', 'Q': -1./3., },
    'Upsilon(2S)': {'tex': r"\Upsilon(2S)\to", 'V': 'Upsilon(2S)', 'Q': -1./3., },
    'Upsilon(3S)': {'tex': r"\Upsilon(3S)\to", 'V': 'Upsilon(3S)', 'Q': -1./3., },
 }

_tex = {'ee': r'e^+e^-', 'mumu': r'\mu^+\mu^-', 'tautau': r'\tau^+\tau^-', 
    'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-',
    'emu,mue': r'e^\pm\mu^\mp', 'etau,taue': r'e^\pm\tau^\mp','mutau,taumu': r'\mu^\pm\tau^\mp',
    'mue,emu': r'e^\pm\mu^\mp', 'taue,etau': r'e^\pm\tau^\mp','taumu,mutau': r'\mu^\pm\tau^\mp'
    }


def _define_obs_V_ll(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $V\to \ell^+\ell^-$ :: $' + _process_tex + r"$"
    _obs_name = "BR("+_hadr[M]['V']+"->"+''.join(ll)+")"
    _obs = Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $"+_process_tex+r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name

def _define_obs_V_ll_ratio(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $V\to \ell^+\ell^-$ :: $' + _process_tex + r"$"
    _obs_name = "R("+_hadr[M]['V']+"->"+''.join(ll)+")"
    _obs = Observable(_obs_name)
    _obs.set_description(r"Ratio of $"+_process_tex+r"$ over BR$(V\to ee)$")
    _obs.tex = r"$\text{R}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name


for M in _hadr:
    for (l1,l2) in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
        _obs_name = _define_obs_V_ll(M, (l1,l2))
        Prediction(_obs_name, Vll_br_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_ratio(M, (l1,l2))
        Prediction(_obs_name, Vll_ratio_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
    for (l1,l2) in [('e','mu'), ('e','tau'), ('mu','tau')]:
        # Combined l1+ l2- + l2+ l1- lepton flavour violating decays
        _obs_name = _define_obs_V_ll(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Vll_br_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_ratio(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Vll_ratio_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))

        _obs_name = _define_obs_V_ll(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Vll_br_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_ratio(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Vll_ratio_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
    for ll in ['e', 'mu', 'tau']:
        _obs_name = _define_obs_V_ll(M, (ll,ll))
        Prediction(_obs_name, Vll_br_func(_hadr[M]['V'], _hadr[M]['Q'], ll,ll))
 



from flavio.math.integrate import nintegrate
from flavio.physics.zdecays.smeftew import gV_SM, gA_SM, _QN
from flavio.physics.common import add_dict
from flavio.classes import Observable, Prediction
from flavio.physics import ckm as ckm_flavio
from . import partondist
import wilson
from numpy import pi, sqrt
import numpy as np
from flavio.config import config

def F_qqll_SM(q, Xq, l, Xl, s, par):
    r"""SM $Z$ and $\gamma$ contribution to the $\bar q q\to \ell^+\ell^-$
    amplitude."""
    # parameters
    Ql = -1
    mZ = par['m_Z']
    GammaZ = 1 / par['tau_Z']
    s2w = par['s2w']
    aEW = par['alpha_e']
    e2 = 4 * pi * aEW
    g2=e2/s2w * 0.9608 # correction factor from running g^2 to 1TeV
    gp2=e2/(1-s2w) * 1.0272 # correction factor from running g'^2 to 1TeV
    g_cw = sqrt(g2+gp2) # g over cw
    s2w = gp2/(g2+gp2)
    par = par.copy()
    par['s2w'] = s2w
    # SM couplings
    Qq = _QN[q]['Q']
    gVq = g_cw * gV_SM(q, par)
    gVl = g_cw * gV_SM(l, par)
    gAq = g_cw * gA_SM(q, par)
    gAl = g_cw * gA_SM(l, par)
    if Xq == 'L':
        gZq = gVq + gAq
    elif Xq == 'R':
        gZq = gVq - gAq
    if Xl == 'L':
        gZl = gVl + gAl
    elif Xl == 'R':
        gZl = gVl - gAl
    # SM contribution
    return g2*s2w * Qq * Ql / s + gZq * gZl / (s - mZ**2 + 1j * mZ * GammaZ)


def wceff_qqll_sm(s, par):
    E = np.einsum('ij,kl->ijkl', np.eye(3), np.eye(3))
    wc = {}
    for Xl in 'LR':
        for Xq in 'LR':
            for q in 'ud':
                wc['CV{}{}_e{}'.format(Xl, Xq, q)] = F_qqll_SM(q, Xq, 'e', Xl, s, par) * E
    wc['CSRL_ed'] = wc['CSRR_ed'] = wc['CTRR_ed'] = np.zeros((3, 3, 3, 3))
    wc['CSRL_eu'] = wc['CSRR_eu'] = wc['CTRR_eu'] = np.zeros((3, 3, 3, 3))
    return wc


def wceff_qqll_np(wc_obj, par, scale):
    r"""Returns Wilson coefficients of the effective Lagrangian
    $$\mathcal{L} = \sum_{q=u,d} C_{eq}^{\Gamma_1\Gamma_2}
    (\bar e_i\Gamma_1 e_j)(\bar q_k\Gamma_2 q_l)$$
    as a dictionary of arrays with shape (3, 3, 3, 3) corresponding to ijkl.
    """
    # get the dictionary
    wcxf_dict = wc_obj.get_wcxf(sector='dB=dL=0', scale=scale, par=par,
                                eft='SMEFT', basis='Warsaw up').dict
    # go to redundant basis, C has all the smeft coefficients
    C = wilson.util.smeftutil.wcxf2arrays_symmetrized(wcxf_dict)

    wc = {}
    ckm = ckm_flavio.get_ckm(par)
    # match to notation used in the observable
    wc['CVLL_eu'] = C['lq1'] - C['lq3']
    wc['CVLL_ed'] = np.einsum('ijmn,mk,nl->ijkl',C['lq1'] + C['lq3'],np.conjugate(ckm),ckm)
    wc['CVRR_eu'] = C['eu']
    wc['CVRR_ed'] = C['ed']
    wc['CVLR_eu'] = C['lu']
    wc['CVLR_ed'] = C['ld']
    wc['CVRL_eu'] = np.einsum('klij->ijkl', C['qe'])
    wc['CVRL_ed'] = np.einsum('mnij,mk,nl->ijkl', C['qe'],np.conjugate(ckm),ckm)
    wc['CSRL_ed'] = np.einsum('ijkm,ml->ijkl',C['ledq'],ckm)
    wc['CSRR_eu'] = -C['lequ1']
    wc['CTRR_eu'] = -C['lequ3']
    wc['CSRL_eu'] = wc['CSRR_ed'] = wc['CTRR_ed'] = np.zeros((3, 3, 3, 3))

    return wc


# translate quark name to LHAPDF flavour index
pdf_flavor = {'d': 1, 'u': 2, 's': 3, 'c': 4, 'b': 5}
fermion_indices = {
    'u': ('u', 0),
    'c': ('u', 1),
    'd': ('d', 0),
    's': ('d', 1),
    'b': ('d', 2),
    'e': ('e', 0),
    'mu': ('e', 1),
    'tau': ('e', 2),
}


def sigma_qqll_partonic(sh, q1, q2, l, wc_eff):
    r"""Total partonic cross section of $q1 q2 \to \ell^+\ell^-$

    Returns $\sigma$ in units of GeV$^{-2}$

    Parameters:
    - `sh`: partonic center of mass energy in GeV$^2$
    - `q1`, `q2`: initial state quarks
    - `l`: final state leptons
    - `wc_eff`: SM + SMEFT amplitude for the process as dictionary with entries corresponding to different Wilson coefficients
    """
    # sh is partonic!
    q, i1 = fermion_indices[q1]
    _q, i2 = fermion_indices[q2]
    if q != _q:
        raise ValueError("Quarks must have the same charge.")
    _, il = fermion_indices[l]
    S = (
        + abs(wc_eff['CVLL_e{}'.format(q)][il, il, i1, i2])**2
        + abs(wc_eff['CVLR_e{}'.format(q)][il, il, i1, i2])**2
        + abs(wc_eff['CVRL_e{}'.format(q)][il, il, i1, i2])**2
        + abs(wc_eff['CVRR_e{}'.format(q)][il, il, i1, i2])**2
        + 3 / 4 * abs(wc_eff['CSRL_e{}'.format(q)][il, il, i1, i2])**2
        + 3 / 4 * abs(wc_eff['CSRR_e{}'.format(q)][il, il, i1, i2])**2
        + 4 * abs(wc_eff['CTRR_e{}'.format(q)][il, il, i1, i2])**2
        # add contributions from h.c. of the effective Lagrangian, (prst) -> (rpts)
        + 3 / 4 * abs(wc_eff['CSRL_e{}'.format(q)][il, il, i2, i1])**2
        + 3 / 4 * abs(wc_eff['CSRR_e{}'.format(q)][il, il, i2, i1])**2
        + 4 * abs(wc_eff['CTRR_e{}'.format(q)][il, il, i2, i1])**2
    )
    return sh / (144 * pi) * S


def dsigma_dtau_qqll(s, tau, l, Q2, wc_eff, par):
    r"""Differential hadronic cross section of $pp\to \ell^+\ell^-$.

    Returns $\frac{d\sigma}{d\tau}$ in units of GeV$^{-2}$.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `tau`: $\tau$, ratio of partonic to hadronic center of mass energy, dimensionless
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `Q2`: factorization scale squared in GeV$^2$
    - `wc_eff`: SM + SMEFT amplitude for the process as dictionary with entries corresponding to different Wilson coefficients
    - `par`: parameter dictionary
    """
    sigma = 0
    members_par = config['PDF set']['dileptons']['members par']
    member = int(par[members_par])
    plumi = partondist.get_parton_lumi(Q2=Q2, member=member)
    for q1 in 'duscb':
        for q2 in 'duscb':
            ud1, i1 = fermion_indices[q1]
            ud2, i2 = fermion_indices[q2]
            if ud1 != ud2:
                continue  # skip cases where q1 and q2 have different charge
            i1 = pdf_flavor[q1]
            i2 = pdf_flavor[q2]
            sh = tau * s
            y = sigma_qqll_partonic(sh, q1, q2, l, wc_eff)
            if y != 0:  # save time if the contribution is anyway zero
                sigma += 2 * plumi.L(i1, -i2, tau) * y
    return sigma


def sigma_qqll_int(s, qmin, qmax, l, Q2, wc_obj, par, scale, newphys=True):
    r"""Integrated hadronic cross section of $pp\to \ell^+\ell^-$.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `qmin`, `qmax`: bin boundaries in the dileption invariant mass, in GeV
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `Q2`: factorization scale squared in GeV$^2$
    - `wc_obj`: Wilson coefficient object
    - `par`: parameter dictionary
    - `scale`: scale at which the Wilson coefficients are evaluated in GeV
    - `newphys`: boolean, whether to include NP contributions or not
    """

    if newphys:
        wc_eff_np = wceff_qqll_np(wc_obj, par, scale)
    else:
        wc_eff_np = {}
    def f(tau):
        wc_eff = wceff_qqll_sm(tau * s, par)
        if newphys:
            # add NP contribution
            wc_eff = add_dict((wc_eff, wc_eff_np))
        return dsigma_dtau_qqll(s, tau, l, Q2, wc_eff, par)
    return nintegrate(f, qmin**2 / s, qmax**2 / s, epsrel=1e-5)


def R_sigma_qqll_int(s, qmin, qmax, l, wc_obj, par):
    r"""Integrated hadronic cross section of $pp\to \ell^+\ell^-$ normalized
    to its SM value.

    Parameters:
    - `s`: hadronic center of mass energy in GeV$^2$
    - `qmin`, `qmax`: bin boundaries in the dileption invariant mass, in GeV
    - `l`: lepton flavour, should be 'e', 'mu', or 'tau'
    - `wc_obj`: Wilson coefficient object
    - `par`: parameter dictionary
    """
    # Renormalization and factorization scale are fixed in the bin
    scale = (qmin + qmax) / 2
    Q2 = scale**2
    sigma_sm = sigma_qqll_int(s, qmin, qmax, l, Q2, wc_obj, par, scale, newphys=False)
    sigma_np = sigma_qqll_int(s, qmin, qmax, l, Q2, wc_obj, par, scale, newphys=True)
    return sigma_np / sigma_sm


def R_sigma_qqll_int_fct_lhc13(l):
    s = 13000**2
    def f(wc_obj, par, qmin, qmax):
        return R_sigma_qqll_int(s, qmin, qmax, l, wc_obj, par)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for l in _tex:
    _process_tex = r"pp\to " + _tex[l] + r"^+" + _tex[l] + r"^-"
    _process_taxonomy = r'Process :: contact interactions :: $' + _process_tex + r"$"

    _obs_name = "R_13(pp->" + 2 * l + ")"
    _obs = Observable(_obs_name, arguments=['qmin', 'qmax'])
    _obs.set_description(r"Cross section of $" + _process_tex + r"$ at $\sqrt{s}=13$ TeV normalized to the SM contribution")
    _obs.tex = r"$\text{R}_{13}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    Prediction(_obs_name, R_sigma_qqll_int_fct_lhc13(l))



r"""$V\to ll^\prime\gamma$ branching ratio"""

from flavio.classes import Observable, Prediction
from flavio.physics.running import running
import flavio
from flavio.physics.quarkonium.Vll import Vll_br, wc_sector
import numpy as np

meson_quark = { 'J/psi' : 'cc', 
                'psi(2S)': 'cc',
                'Upsilon(1S)': 'bb',
                'Upsilon(2S)': 'bb',
                'Upsilon(3S)': 'bb',
                }

def getWC_lfv(wc_obj,par,V,Q,l1,l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji):
    # renormalization scale
    scale = flavio.config['renormalization scale'][V]
    # Wilson coefficients
    wc = wc_obj.get_wc(wc_sector[(l1,l2)], scale, par)

    alphaem = running.get_alpha_e(par, scale)
    ee=np.sqrt(4.*np.pi*alphaem) 

    mV = par['m_'+V]   
    fV=par['f_'+V] 
    fV_T=par['fT_'+V]
    qq=meson_quark[V]
    # for emu and taue the name of the Wilson coefficient sector agrees with the ordering of leptons in the vector bilinear
    # This is not the case for mutau. Thus distinguish between the two cases here.
    if wc_sector[(l1,l2)]=="mutau":
        ll="taumu"
    else:
        ll=wc_sector[(l1,l2)]
    VR=fV*mV*(wc['CVRR_'+ll+qq] + wc['CVLR_'+qq+ll]) 
    VL=fV*mV*(wc['CVLL_'+ll+qq] + wc['CVLR_'+ll+qq]) 
    TR=fV_T*mV*wc['CTRR_'+l1+l2+qq] - ee*Q *fV*wc['Cgamma_'+l2+l1] 
    TL=(fV_T*mV*wc['CTRR_'+l2+l1+qq] - ee*Q *fV*wc['Cgamma_'+l1+l2]).conjugate()
    SR=2.*mV*fV*(wc['CSRR_'+l1+l2+qq]+wc['CSRL_'+l1+l2+qq])
    SL=2.*mV*fV*(wc['CSRL_'+l2+l1+qq]+wc['CSRR_'+l2+l1+qq]).conjugate()   
    PR=2.*mV*fV*(wc['CSRR_'+l1+l2+qq]-wc['CSRL_'+l1+l2+qq])
    PL=2.*mV*fV*(wc['CSRL_'+l2+l1+qq]-wc['CSRR_'+l2+l1+qq]).conjugate()   
    AR=2.*fV*mV*(wc['CVLR_'+qq+ll]- wc['CVRR_'+ll+qq]) 
    AL=2.*fV*mV*(wc['CVLL_'+ll+qq] - wc['CVLR_'+ll+qq]) 

    StildeR=4*mV**2*fV*CeFFij
    StildeL=4*mV**2*fV*CeFFji.conjugate()
    PtildeR=4*mV**2*fV*CeFFtildeij
    PtildeL=4*mV**2*fV*CeFFtildeji.conjugate()

    if ll==l2+l1: 
    # As we use ll instead of l1+l2 or l2+l1 above when constructing the effective parameters, we have to complex conjugate it, in case indices are reversed.
    # This is not necessary for the others, because there we explicitly defined the flavour indices.
        VL=VL.conjugate()
        VR=VR.conjugate()
        AL=AL.conjugate()
        AR=AR.conjugate()
    return VL,VR,TL,TR,SR,SL,PR,PL,AR,AL,StildeR,StildeL,PtildeR,PtildeL

# we use \mu=y^2 inside the kinematic functions and denote them by F_ instead of G_ in the paper.

def F_A(y):
    mu=y**2
    if mu==0:
        return 2./9
    return (8.-45*mu+36*mu**2+mu**3+6*(mu-6)*mu**2*np.log(mu))/36
def F_S(y):
    mu=y**2
    if mu==0:
        return 1./12
    return (1-6*mu+3*mu**2*(1-2*np.log(mu))+2*mu**3)/12.
def Ftilde_S(y):
    mu=y**2
    if mu==0:
        return 1./40
    return (3.-30.*mu-20.*mu**2+60.*mu**3-15*mu**4+2*mu**5-60.*mu**2*np.log(mu))/120.
def Fhat_P(y):
    mu=y**2
    if mu==0:
        return 0.
    return mu*(-8+8*mu**2-mu**3 - 12*mu*np.log(mu))/12.
def Fhat_S(y):
    mu=y**2
    if mu==0:
        return 1./12
    return Fhat_P(y) + 1./12
def F_PA(y):
    if y==0:
        return 0.
    mu=y**2
    return y*(1.+4*mu-5*mu**2+2*mu*(2+mu)*np.log(mu))/2.
def Ftilde_PA(y):
    if y==0:
        return 0.
    mu=y**2
    return y*(1.+9*mu-9*mu**2-mu**3+6*mu*(1+mu)*np.log(mu))/3.


def Vllgamma_br(wc_obj, par,V,Q, l1,l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji):
    r"""Branching ratio for the lepton-flavour violating leptonic decay J/psi-> l l' \gamma"""
    #####branching ratio obtained from 2207.10913#####
    flavio.citations.register("Calibbi:2022ddo")
    # renormalization scale
    scale = flavio.config['renormalization scale'][V]
    alphaem = running.get_alpha_e(par, scale)
    mV = par['m_'+V]   
    tauV = par['tau_'+V]  
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]

    VL,VR,TL,TR,SR,SL,PR,PL,AR,AL,StildeR,StildeL,PtildeR,PtildeL = getWC_lfv(wc_obj,par,V,Q,l1,l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)
 
    AV=np.abs(AL)**2+np.abs(AR)**2
    SP=np.abs(SL)**2+np.abs(PL)**2+np.abs(SR)**2+np.abs(PR)**2
    SPtilde=np.abs(StildeL)**2+ np.abs(PtildeL)**2 +np.abs(StildeR)**2  +np.abs(PtildeR)**2
    SStilde= (SL*StildeL.conjugate() + SR*StildeR.conjugate()).real
    PPtilde= (PL*PtildeL.conjugate() + PR*PtildeR.conjugate()).imag

   
    if ml1<ml2:
        y=ml2/mV
        I_AP=-(AL*PR.conjugate()+AR*PL.conjugate()).real
        I_APtilde=-(AL*PtildeR.conjugate()+AR*PtildeL.conjugate()).imag
    elif ml2<ml1:
        y=ml1/mV
        I_AP=(AL*PL.conjugate()+AR*PR.conjugate()).real
        I_APtilde=(AL*PtildeL.conjugate()+AR*PtildeR.conjugate()).imag
    else:
        print("The case of non-hierarchical masses is not implemented.")

    prefactor=alphaem*Q**2*tauV*mV/(192*np.pi**2)
    
    return prefactor*(AV * F_A(y) + SP*F_S(y) + I_AP*F_PA(y) + SPtilde *Ftilde_S(y) + SStilde * Fhat_S(y) + PPtilde * Fhat_P(y) +I_APtilde * Ftilde_PA(y))


def Vllgamma_br_func(V, Q, l1, l2):
    def fct(wc_obj, par,CeFFij=0,CeFFji=0,CeFFtildeij=0,CeFFtildeji=0):
        return Vllgamma_br(wc_obj, par, V, Q, l1, l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)
    return fct


def Vllgamma_br_comb_func(V, Q, l1, l2):
    def fct(wc_obj, par,CeFFij=0,CeFFji=0,CeFFtildeij=0,CeFFtildeji=0):
        return Vllgamma_br(wc_obj, par, V, Q, l1, l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)+ Vllgamma_br(wc_obj, par, V, Q, l2, l1,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)
    return fct

def Vllgamma_ratio_func(V, Q, l1, l2):
    def fct(wc_obj, par,CeFFij=0,CeFFji=0,CeFFtildeij=0,CeFFtildeji=0):
        BRee=Vll_br(wc_obj,par,V,Q,'e','e')
        return Vllgamma_br(wc_obj, par, V, Q, l1, l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)/BRee
    return fct


def Vllgamma_ratio_comb_func(V, Q, l1, l2):
    def fct(wc_obj, par,CeFFij=0,CeFFji=0,CeFFtildeij=0,CeFFtildeji=0):
        BRee=Vll_br(wc_obj,par,V,Q,'e','e')
        return (Vllgamma_br(wc_obj, par, V, Q, l1, l2,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji)+ Vllgamma_br(wc_obj, par, V, Q, l2, l1,CeFFij,CeFFji,CeFFtildeij,CeFFtildeji))/BRee
    return fct



# Observable and Prediction instances
_hadr = { 
    'J/psi': {'tex': r"J/\psi\to", 'V': 'J/psi', 'Q': 2./3., },
    'psi(2S)': {'tex': r"\psi(2S)\to", 'V': 'psi(2S)', 'Q': 2./3., }, 
    'Upsilon(1S)': {'tex': r"\Upsilon(1S)\to", 'V': 'Upsilon(1S)', 'Q': -1./3., },
    'Upsilon(2S)': {'tex': r"\Upsilon(2S)\to", 'V': 'Upsilon(2S)', 'Q': -1./3., },
    'Upsilon(3S)': {'tex': r"\Upsilon(3S)\to", 'V': 'Upsilon(3S)', 'Q': -1./3., },
 }

_tex = {'ee': r'e^+e^-', 'mumu': r'\mu^+\mu^-', 'tautau': r'\tau^+\tau^-', 
    'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-',
    'emu,mue': r'e^\pm\mu^\mp', 'etau,taue': r'e^\pm\tau^\mp','mutau,taumu': r'\mu^\pm\tau^\mp',
    'mue,emu': r'e^\pm\mu^\mp', 'taue,etau': r'e^\pm\tau^\mp','taumu,mutau': r'\mu^\pm\tau^\mp'
    }


def _define_obs_V_ll_gamma(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $V\to \ell^+\ell^-\gamma$ :: $' + _process_tex + r"$"
    _obs_name = "BR("+_hadr[M]['V']+"->"+''.join(ll)+"gamma)"
    _obs = Observable(_obs_name,arguments=["CeFFij","CeFFji","CeFFtildeij","CeFFtildeji"])
    _obs.set_description(r"Branching ratio of $"+_process_tex+r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name

def _define_obs_V_ll_gamma_ratio(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $V\to \ell^+\ell^-\gamma$ :: $' + _process_tex + r"$"
    _obs_name = "R("+_hadr[M]['V']+"->"+''.join(ll)+"gamma)"
    _obs = Observable(_obs_name,arguments=["CeFFij","CeFFji","CeFFtildeij","CeFFtildeji"])
    _obs.set_description(r"Ratio of $"+_process_tex+r"$ over BR$(V\to ee)$")
    _obs.tex = r"$\text{R}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name


for M in _hadr:
    for (l1,l2) in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
        _obs_name = _define_obs_V_ll_gamma(M, (l1,l2))
        Prediction(_obs_name, Vllgamma_br_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_gamma_ratio(M, (l1,l2))
        Prediction(_obs_name, Vllgamma_ratio_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))

    for (l1,l2) in [('e','mu'), ('e','tau'), ('mu','tau')]:
        # Combined l1+ l2- + l2+ l1- lepton flavour violating decays
        _obs_name = _define_obs_V_ll_gamma(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Vllgamma_br_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_gamma_ratio(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Vllgamma_ratio_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        
        _obs_name = _define_obs_V_ll_gamma(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Vllgamma_br_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
        _obs_name = _define_obs_V_ll_gamma_ratio(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Vllgamma_ratio_comb_func(_hadr[M]['V'], _hadr[M]['Q'], l1,l2))
 



"""Functions for parton distributions."""


import parton
from functools import lru_cache
import scipy.interpolate
import numpy as np
from flavio.config import config


@lru_cache()
def get_pdf(name, member=0, download=False):
    try:
        pdf = parton.PDF(name, member)
    except ValueError as e:
        if "Data file" in str(e):
            raise Exception(f"PDF set '{name}' does not provide member '{member}'")
        elif download is True:
            dir = parton.io.data_dir()
            parton.io.download_pdfset(name, dir)
            pdf = parton.PDF(name, member)
        else:
            raise Exception(f"PDF set '{name}' not available. Please install the PDF set using the command \"python3 -m parton install '{name}'\"")
    return pdf

@lru_cache(maxsize=None)
def get_parton_lumi(Q2, member=0):
    pdfset = config['PDF set']['dileptons']['name']
    pdf = get_pdf(pdfset, member)
    return parton.PLumi(pdf, Q2=Q2)



import numpy as np
import flavio
from flavio.physics.bdecays.formfactors.common import z

def pole(ff, mres, q2):
    if mres == 0 or mres is None:
        return 1
    return 1/(1-q2/mres**2)

# the following dict maps transitions to mesons. Note that it doesn't really
# matter whether the charged or neutral B/K/pi are used here. We don't
# distinguish between charged and neutral form factors anyway.
process_dict = {}
process_dict['B->pi'] =   {'B': 'B+', 'P': 'pi0'}


def param_fplusT(mB, mP, b_i, q2, t0=None):
    # b_i[0] is f(q^2 = 0), b_i[1:] should be multiplied with b_i[0]
    Z = z(mB, mP, q2, t0)
    Z0 = z(mB, mP, 0, t0)
    n = len(b_i)
    k = np.arange(1, n, 1)
    return b_i[0]*(1+( b_i[1:] * (Z**k-Z0**k - (-1.)**(k - n) * k/n * (Z**n-Z0**n)) ).sum())

def param_f0(mB, mP, b_i, q2, t0=None):
    # b_i[0] is f(q^2 = 0), b_i[1:] should be multiplied with b_i[0]
    Z = z(mB, mP, q2, t0)
    Z0 = z(mB, mP, 0, t0)
    n = len(b_i)
    k = np.arange(1 ,n, 1)
    return b_i[0]*(1+( b_i[1:] * (Z**k -Z0**k)).sum())

def ff(process, q2, par, n=4, t0=None):
    r"""Central value of $B\to P$ form factors in the standard convention
    and BCL parametrization from arXiv:2102.07233.

    The standard convention defines the form factors $f_+$, $f_0$, and $f_T$.
    """
    flavio.citations.register("Leljak:2021vte")
    pd = process_dict[process]
    mpl = par[process + ' BCL LMVD m+']
    m0 = par[process + ' BCL LMVD m0']
    mB = par[process + ' BCL LMVD m_' + pd['B']]
    mP = par[process + ' BCL LMVD m_' + pd['P']]
    ff = {}
    b={}
    for i in ['f+', 'fT']:
        b[i] = [par[process + ' BCL' + ' f' + f'_{i[1:]}(0)']]+ [ par[process + ' BCL' + ' b' f'_{i[1:]}^{j}'] for j in range(1, n)]
    # evaluate FFs
    ff['f+'] = pole('f+', mpl, q2) * param_fplusT(mB, mP, b['f+'], q2, t0)
    ff['fT'] = pole('fT', mpl, q2) * param_fplusT(mB, mP, b['fT'], q2, t0)

    # f0 is modified
    b['f0'] = [par[process + ' BCL' + ' f' + f'_+(0)']]+ [ par[process + ' BCL' + ' b' + f'_0^{j}'] for j in range(1, n+1)] # note the +1
    ff['f0'] = pole('f0', m0, q2) * param_f0(mB, mP, b['f0'], q2, t0)
    return ff



import yaml
import pkgutil
import numpy as np
from flavio.classes import Parameter
from flavio.statistics.probability import MultivariateNormalDistribution

def load_parameters(filename, constraints):
    f = pkgutil.get_data('flavio.physics', filename)
    ff_dict = yaml.safe_load(f)
    dict_name = list(ff_dict.keys())[0]
    observables = ff_dict[dict_name]['observables']
    central_values = np.asarray(ff_dict[dict_name]['medians'])
    covariance = np.asarray(ff_dict[dict_name]['covariance'])
    observables = [o.replace('::', '') for o in observables]
    observables_renamed = []
    for o in observables:
        try:
            index = o.index('f')
        except:
            try:
                index = o.index('b')
            except:
                raise ValueError('No f or b in observable name')
        observables_renamed.append(o[:index]+ ' BCL ' + o[index:])

    for parameter_name in observables_renamed:
        try: # check if parameter object already exists
            p = Parameter[parameter_name]
        except: # otherwise, create a new one
            p = Parameter(parameter_name)
        else: # if parameter exists, remove existing constraints
            constraints.remove_constraint(parameter_name)
    if not np.allclose(covariance, covariance.T):
        # if the covariance is not symmetric, it is assumed that only the values above the diagonal are present.
        # then: M -> M + M^T - diag(M)
        covariance = covariance + covariance.T - np.diag(np.diag(covariance))
    constraints.add_constraint(observables_renamed,
            MultivariateNormalDistribution(central_value=central_values, covariance=covariance) )



from flavio.physics.bdecays.formfactors.lambdab_32 import QuarkModel_MCN, LatticeQCD
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.config import config


processes = ['Lambdab->Lambda(1520)',]


def ff_function(function, process):
    return lambda wc_obj, par_dict, q2: function(process, q2, par_dict)


for p in processes:
    quantity = p + ' form factor'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2'])
    a.set_description('Hadronic form factor for the ' + p + ' transition')

    # MCN quark model approach as in arXiv:1108.6129
    iname = p + ' MCN'
    i = Implementation(name=iname, quantity=quantity,
                       function=ff_function(QuarkModel_MCN.ff_equiv, p))
    i.set_description("Form factors calculated using the full quark model wave function with the full relativistic form of the quark current")

    # Lattice QCD prediction arXiv:2009.09313
    iname = p + ' LatticeQCD'
    i = Implementation(name=iname, quantity=quantity,
                       function=ff_function(LatticeQCD.ff_equiv, p))
    i.set_description("Form factors from lattice QCD valid in $q^2 \in [16; 16.8] GeV^2$")



import unittest
import flavio
from wilson import Wilson
from .Vll import *

par = flavio.default_parameters.get_central_all()

### implement test
class TestVll(unittest.TestCase):
    def test_sm(self):
        par=flavio.default_parameters.get_central_all()
        Gamma_Jpsi_ee = 5.637e-6 # 2005.01845 abstract 5.637(49)e-6 
        br=Gamma_Jpsi_ee*par['tau_J/psi']
        self.assertAlmostEqual(flavio.sm_prediction('BR(J/psi->ee)'), br,delta=0.1*br) # Our leading order result is not expected to exactly agree with 2005.01845 and thus we only require to agree within 10%

    def test_np(self):
        wc,br=Wilson({'CVRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.01788
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->mue)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->mue)',wc), flavio.np_prediction('BR(J/psi->mue)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)

        wc,br=Wilson({'CTRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.1312
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->mue)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->mue)',wc), flavio.np_prediction('BR(J/psi->mue)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)

        wc,br=Wilson({'CVRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.009387
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->taue)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->taue)',wc), flavio.np_prediction('BR(J/psi->taue)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)
        
        wc,br=Wilson({'CTRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.09791
        self.assertAlmostEqual(flavio.np_prediction('BR(J/psi->taue)',wc), br,delta=0.01*br)
        self.assertAlmostEqual(flavio.np_prediction('R(J/psi->taue)',wc), flavio.np_prediction('BR(J/psi->taue)',wc)/flavio.np_prediction('BR(J/psi->ee)',wc),delta=0.001*br)



r"""$P\to ll^\prime$ branching ratio"""

from flavio.classes import Observable, Prediction
import flavio
from flavio.physics.running import running
from flavio.physics.common import lambda_K
from flavio.physics.quarkonium.Vll import wc_sector
import numpy as np

meson_quark = { 'eta_c(1S)': 'cc', 
                'eta_b(1S)': 'bb',
                }


def getS_lfv(wc_obj,par,P,l1,l2,CeGGij,CeGGji):
    # renormalization scale
    scale = flavio.config['renormalization scale'][P]
    alphas = running.get_alpha_s(par, scale)
    # Wilson coefficients
    wc = wc_obj.get_wc(wc_sector[(l1,l2)], scale, par)

    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    mP = par['m_'+P]   
    fP=par['f_'+P] 
    aP=par['a_'+P]
    hP= mP**2 * fP -aP
    qq=meson_quark[P]
    mq=par['m_'+qq[0]]
    # for emu and taue the name of the Wilson coefficient sector agrees with the ordering of leptons in the vector bilinear
    # This is not the case for mutau. Thus distinguish between the two cases here.
    if wc_sector[(l1,l2)]=="mutau":
        ll="taumu"
    else:
        ll=wc_sector[(l1,l2)]
    aPfac=1j*aP*4.*np.pi/alphas
    SR= aPfac * CeGGij +  (hP/(4*mq))*(wc['CSRR_'+l1+l2+qq]-wc['CSRL_'+l1+l2+qq])  -(fP/2.)*(ml1*(wc['CVRR_'+ll+qq] - wc['CVLR_'+qq+ll])+ml2* (wc['CVLR_'+ll+qq] -wc['CVLL_'+ll+qq] ))  
    SL= aPfac * CeGGji + (hP/(4*mq))*(wc['CSRL_'+l2+l1+qq]-wc['CSRR_'+l2+l1+qq]).conjugate() -(fP/2.)*(ml2*(wc['CVRR_'+ll+qq] - wc['CVLR_'+qq+ll])+ml1* (wc['CVLR_'+ll+qq] -wc['CVLL_'+ll+qq] )) 
    return SL,SR


def Pll_br(wc_obj, par,P, l1,l2,CeGGij,CeGGji):
    r"""Branching ratio for the lepton-flavour violating leptonic decay P -> l l' """
    #####branching ratio obtained from 2207.10913#####
    flavio.citations.register("Calibbi:2022ddo")
    mP = par['m_'+P]   
    tauP = par['tau_'+P]  
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    y1=ml1/mP
    y2=ml2/mP
    y1s=y1**2
    y2s=y2**2
    SL,SR = getS_lfv(wc_obj,par,P,l1,l2,CeGGij,CeGGji)
    return  tauP*mP/(16.*np.pi) * np.sqrt(lambda_K(1,y1s,y2s)) * ((1-y1s-y2s)*(np.abs(SL)**2+np.abs(SR)**2) -4*y1*y2 *(SL*SR.conjugate()).real)


def Pll_br_func(P,  l1, l2):
    def fct(wc_obj, par,CeGGij=0,CeGGji=0):
        return Pll_br(wc_obj, par, P,  l1, l2,CeGGij,CeGGji)
    return fct


def Pll_br_comb_func(P,  l1, l2):
    def fct(wc_obj, par,CeGGij=0,CeGGji=0):
        return Pll_br(wc_obj, par, P,  l1, l2,CeGGij,CeGGji)+ Pll_br(wc_obj, par, P,  l2, l1,CeGGij,CeGGji)
    return fct


# Observable and Prediction instances
_hadr = { 
    'eta_c(1S)': {'tex': r"\eta_c(1S)\to", 'P': 'eta_c(1S)' },
    'eta_b(1S)': {'tex': r"\eta_b(1S)\to", 'P': 'eta_b(1S)' },
 }

_tex = {'ee': r'e^+e^-', 'mumu': r'\mu^+\mu^-', 'tautau': r'\tau^+\tau^-', 
    'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-',
    'emu,mue': r'e^\pm\mu^\mp', 'etau,taue': r'e^\pm\tau^\mp','mutau,taumu': r'\mu^\pm\tau^\mp',
    'mue,emu': r'e^\pm\mu^\mp', 'taue,etau': r'e^\pm\tau^\mp','taumu,mutau': r'\mu^\pm\tau^\mp'
    }


def _define_obs_P_ll(M, ll):
    _process_tex = _hadr[M]['tex']+_tex[''.join(ll)]
    _process_taxonomy = r'Process :: quarkonium lepton decays :: $P\to \ell^+\ell^-$ :: $' + _process_tex + r"$"
    _obs_name = "BR("+_hadr[M]['P']+"->"+''.join(ll)+")"
    _obs = Observable(_obs_name,arguments=["CeGGij","CeGGji"])
    _obs.set_description(r"Branching ratio of $"+_process_tex+r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex+r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs_name


for M in _hadr:
    for (l1,l2) in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
        _obs_name = _define_obs_P_ll(M, (l1,l2))
        Prediction(_obs_name, Pll_br_func(_hadr[M]['P'], l1, l2))
    for ll in [('e','mu'), ('e','tau'), ('mu','tau')]:
        # Combined l1+ l2- + l2+ l1- lepton flavour violating decays
        _obs_name = _define_obs_P_ll(M, ('{0}{1},{1}{0}'.format(l1,l2),))
        Prediction(_obs_name, Pll_br_comb_func(_hadr[M]['P'], l1,l2))
        _obs_name = _define_obs_P_ll(M, ('{1}{0},{0}{1}'.format(l1,l2),))
        Prediction(_obs_name, Pll_br_comb_func(_hadr[M]['P'], l1,l2))
 



import unittest
import flavio
from wilson import Wilson
from .Sll import *

### implement test
class TestSll(unittest.TestCase):
    def test_np(self):
        wc,br=Wilson({'CSRR_muecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.0014399
        self.assertAlmostEqual(flavio.np_prediction('BR(chi_c0(1P)->mue)',wc), br,delta=0.001*br)

        wc,br=Wilson({'CSRR_tauecc' : 1e-2},scale=2.,eft='WET',basis='flavio'),0.00076716
        self.assertAlmostEqual(flavio.np_prediction('BR(chi_c0(1P)->taue)',wc), br,delta=0.001*br)



r"""Functions for $B\to P$ form factors"""


from . import btop
from . import bcl
from . import bcl_lmvd
from . import bcl_parameters_lmvd
from . import bcl_parameters
from . import bsz_parameters
from . import isgurwise



import unittest
import flavio
from flavio.physics.dileptons import ppll, partondist
from wilson import wcxf
import numpy as np
from flavio.config import config

par = flavio.default_parameters.get_central_all()
par2 = par.copy()
par2['s2w'] = 0.221841          # correction factor to match the madgraph input scheme
par2['alpha_e'] = 0.00772585    # correction factor to match the madgraph input scheme
GeVtopb = 0.3894*10**9

pdf_set_for_tests = 'NNPDF30_nnlo_as_0118'
config['PDF set']['dileptons']['name'] = pdf_set_for_tests

def uses_pdf(func):
    def inner(*args, **kwargs):
        partondist.get_pdf(pdf_set_for_tests, download=True)
        return func(*args, **kwargs)
    return inner

class Test_QQLL(unittest.TestCase):
    @uses_pdf
    def test_sm_spectrum(self):
        r"""Compare the SM $m_{\ell\ell}$ spectrum from flavio to madgraph prediction in 10 bins from 200 GeV to 2 TeV
            MG events simulated with NNPDF30_nnlo_as_0118 PDFs
            2% acceptable error"""

        spectrum_mg = np.asarray([2.66250518e-02, 8.76624628e-03, 2.92981721e-03, 9.64842852e-04, 3.12854454e-04, 9.77018348e-05, 2.84022912e-05, 8.08026977e-06, 2.02974822e-06])
        # Number of SM events generated in MG [60169., 25284., 10791., 25514., 52059., 20750., 42577., 45628., 14659.]

        s = 13e3**2
        logbins = np.linspace(np.log10(200),np.log10(1800),10)
        bins = 10**logbins
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            width = bins[i+1]-bins[i]
            spectrum = ppll.sigma_qqll_int(s, bins[i], bins[i+1], 'mu', center**2, 0, par2, 0, newphys=False)*GeVtopb/width
            err = (spectrum-spectrum_mg[i])/spectrum_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqll_sm(self):
        """Test the SM pediction for the R ratio from 1 to 2 TeV"""

        R = flavio.sm_prediction('R_13(pp->mumu)', 1e3, 2e3)
        self.assertEqual(R, 1,msg=f'SM prediction for R ratio: {R} (has to be 1)')

    @uses_pdf
    def test_qqll_np_sm(self):
        r"""Test the R ratio with presence of wilson coefficients set to zero"""

        wc_obj = flavio.WilsonCoefficients()
        R = ppll.R_sigma_qqll_int(13e3**2, 1300, 1800, 'mu', wc_obj, par)
        self.assertEqual(R, 1,msg=f'SM prediction for R ratio: {R} (has to be 1)')

    @uses_pdf
    def test_qqll_np_vector(self):
        r"""Testthe R ratio for one specific vector WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118, SMEFTsim_general_alphaScheme_UFO
            2% acceptable error"""

        R_mg = np.asarray([1.1998640111246237, 1.6266639004553087, 2.4701438261611193])
        # Number of NP events generated in MG [37802, 24730, 16780]
        # Number of SM events generated in MG [58147, 23848, 10473]

        bins = np.asarray([1000., 1200., 1400., 1600.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'lq1_2211': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = ppll.R_sigma_qqll_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqll_np_scalar(self):
        r"""Test the R ratio for one specific scalar WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118, SMEFTsim_general_alphaScheme_UFO
            2% acceptable error"""

        R_mg = np.asarray([1.0159054527417566, 1.0257049317291311, 1.0386032816907145])
        # Number of NP events generated in MG [38339, 25414, 16752]
        # Number of SM events generated in MG [58147, 23848, 10473]

        bins = np.asarray([1000., 1200., 1400., 1600.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'ledq_2223': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = ppll.R_sigma_qqll_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')

    @uses_pdf
    def test_qqll_np_tensor(self):
        r"""Test the R ratio for one specific tensor WC
            comparing to results from MadGraph using NNPDF30_nnlo_as_0118, SMEFTsim_general_alphaScheme_UFO
            2% acceptable error"""

        R_mg = np.asarray([2.1352791354283926, 3.0789137670371165, 4.609226551698083])
        # Number of NP events generated in MG [32345, 24282, 18505]
        # Number of SM events generated in MG [58147, 23848, 10473]

        bins = np.asarray([1000., 1200., 1400., 1600.])
        nbins = len(bins)-1

        for i in range(nbins):
            center = 0.5*(bins[i]+bins[i+1])
            wc = wcxf.WC('SMEFT', 'Warsaw up', center, {'lequ3_2212': 1e-7})
            wc_obj = flavio.WilsonCoefficients()
            wc_obj.set_initial_wcxf(wc)
            R = ppll.R_sigma_qqll_int(13e3**2, bins[i], bins[i+1], 'mu', wc_obj, par2)
            err = (R-R_mg[i])/R_mg[i]
            self.assertAlmostEqual(err,0,delta=0.02,msg=f'error in bin {i}: {err}')
